# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_trick_level.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetTrickLevel(object):
    def setupUi(self, PresetTrickLevel):
        if not PresetTrickLevel.objectName():
            PresetTrickLevel.setObjectName(u"PresetTrickLevel")
        PresetTrickLevel.resize(539, 516)
        self.centralWidget = QWidget(PresetTrickLevel)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.trick_level_scroll = QScrollArea(self.centralWidget)
        self.trick_level_scroll.setObjectName(u"trick_level_scroll")
        self.trick_level_scroll.setFrameShape(QFrame.NoFrame)
        self.trick_level_scroll.setFrameShadow(QFrame.Plain)
        self.trick_level_scroll.setWidgetResizable(True)
        self.trick_level_scroll_contents = QWidget()
        self.trick_level_scroll_contents.setObjectName(u"trick_level_scroll_contents")
        self.trick_level_scroll_contents.setGeometry(QRect(0, 0, 525, 625))
        self.trick_level_layout = QVBoxLayout(self.trick_level_scroll_contents)
        self.trick_level_layout.setSpacing(6)
        self.trick_level_layout.setContentsMargins(11, 11, 11, 11)
        self.trick_level_layout.setObjectName(u"trick_level_layout")
        self.trick_level_layout.setContentsMargins(4, 8, 4, 0)
        self.logic_description_label = QLabel(self.trick_level_scroll_contents)
        self.logic_description_label.setObjectName(u"logic_description_label")
        self.logic_description_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.logic_description_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.logic_description_label)

        self.trick_level_line_1 = QFrame(self.trick_level_scroll_contents)
        self.trick_level_line_1.setObjectName(u"trick_level_line_1")
        self.trick_level_line_1.setFrameShape(QFrame.HLine)
        self.trick_level_line_1.setFrameShadow(QFrame.Sunken)

        self.trick_level_layout.addWidget(self.trick_level_line_1)

        self.dangerous_layout = QHBoxLayout()
        self.dangerous_layout.setSpacing(6)
        self.dangerous_layout.setObjectName(u"dangerous_layout")
        self.dangerous_label = QLabel(self.trick_level_scroll_contents)
        self.dangerous_label.setObjectName(u"dangerous_label")

        self.dangerous_layout.addWidget(self.dangerous_label)

        self.dangerous_combo = QComboBox(self.trick_level_scroll_contents)
        self.dangerous_combo.addItem("")
        self.dangerous_combo.addItem("")
        self.dangerous_combo.setObjectName(u"dangerous_combo")

        self.dangerous_layout.addWidget(self.dangerous_combo)


        self.trick_level_layout.addLayout(self.dangerous_layout)

        self.dangerous_description = QLabel(self.trick_level_scroll_contents)
        self.dangerous_description.setObjectName(u"dangerous_description")
        self.dangerous_description.setWordWrap(True)

        self.trick_level_layout.addWidget(self.dangerous_description)

        self.trick_level_line_2 = QFrame(self.trick_level_scroll_contents)
        self.trick_level_line_2.setObjectName(u"trick_level_line_2")
        self.trick_level_line_2.setFrameShape(QFrame.HLine)
        self.trick_level_line_2.setFrameShadow(QFrame.Sunken)

        self.trick_level_layout.addWidget(self.trick_level_line_2)

        self.trick_level_minimal_logic_check = QCheckBox(self.trick_level_scroll_contents)
        self.trick_level_minimal_logic_check.setObjectName(u"trick_level_minimal_logic_check")

        self.trick_level_layout.addWidget(self.trick_level_minimal_logic_check)

        self.trick_level_minimal_logic_label = QLabel(self.trick_level_scroll_contents)
        self.trick_level_minimal_logic_label.setObjectName(u"trick_level_minimal_logic_label")
        self.trick_level_minimal_logic_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.trick_level_minimal_logic_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.trick_level_minimal_logic_label)

        self.underwater_abuse_check = QCheckBox(self.trick_level_scroll_contents)
        self.underwater_abuse_check.setObjectName(u"underwater_abuse_check")

        self.trick_level_layout.addWidget(self.underwater_abuse_check)

        self.underwater_abuse_label = QLabel(self.trick_level_scroll_contents)
        self.underwater_abuse_label.setObjectName(u"underwater_abuse_label")
        self.underwater_abuse_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.underwater_abuse_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.underwater_abuse_label)

        self.trick_level_line_3 = QFrame(self.trick_level_scroll_contents)
        self.trick_level_line_3.setObjectName(u"trick_level_line_3")
        self.trick_level_line_3.setFrameShape(QFrame.HLine)
        self.trick_level_line_3.setFrameShadow(QFrame.Sunken)

        self.trick_level_layout.addWidget(self.trick_level_line_3)

        self.trick_level_help_label = QLabel(self.trick_level_scroll_contents)
        self.trick_level_help_label.setObjectName(u"trick_level_help_label")
        self.trick_level_help_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.trick_level_help_label)

        self.trick_level_scroll.setWidget(self.trick_level_scroll_contents)

        self.verticalLayout.addWidget(self.trick_level_scroll)

        PresetTrickLevel.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetTrickLevel)

        QMetaObject.connectSlotsByName(PresetTrickLevel)
    # setupUi

    def retranslateUi(self, PresetTrickLevel):
        PresetTrickLevel.setWindowTitle(QCoreApplication.translate("PresetTrickLevel", u"Trick Level", None))
        self.logic_description_label.setText(QCoreApplication.translate("PresetTrickLevel", u"<html><head/><body><p align=\"justify\">Randovania has rules in place which guarantees that the game is completable regardless of the modifications made to the game. Here you can also configure which kind of game knowledge or skill it expects you to have, allowing for even more varied games.</p><p align=\"justify\">No matter the level, it is always possible to softlock when you enter a room or area that you're unable to leave. For example, vanilla beam rooms without the necessary beam to escape, Dark World without Light Beam/Anihhilator Beam, Torvus Bog without Super Missile.</p><p align=\"justify\">However, it may be <span style=\" font-style:italic;\">necessary</span> to enter Dark World without a way to escape if that item is located in the Dark World.</p></body></html>", None))
        self.dangerous_label.setText(QCoreApplication.translate("PresetTrickLevel", u"Dangerous actions:", None))
        self.dangerous_combo.setItemText(0, QCoreApplication.translate("PresetTrickLevel", u"Randomly", None))
        self.dangerous_combo.setItemText(1, QCoreApplication.translate("PresetTrickLevel", u"Last Resort", None))

        self.dangerous_description.setText(QCoreApplication.translate("PresetTrickLevel", u"<html><head/><body><p>A dangerous action is the act of moving past a lock without the appropriate items needed to head backwards, or doing an action that can only be done once.</p><p><span style=\" font-weight:600;\">Randomly</span>: Dangerous actions might be required by logic.</p><p><span style=\" font-weight:600;\">Last Resort</span>: Only allows dangerous actions to be required if no other option is available for progression.<br/>Warning: Due to how item placement works, certain locations will have progression extremely less often or even never.</p></body></html>", None))
        self.trick_level_minimal_logic_check.setText(QCoreApplication.translate("PresetTrickLevel", u"Use minimal logic instead", None))
        self.trick_level_minimal_logic_label.setText(QCoreApplication.translate("PresetTrickLevel", u"<html><head/><body><p>{game_specific_text}</p><p>There are no guarantees that a seed will be possible in this case.</p></body></html>", None))
        self.underwater_abuse_check.setText(QCoreApplication.translate("PresetTrickLevel", u"Allow abuse of underwater movement without Gravity Suit", None))
        self.underwater_abuse_label.setText(QCoreApplication.translate("PresetTrickLevel", u"<html><head/><body><p>Movement while not having Gravity Suit has different physics, which causes different item requirements. <a href=\"resource-details://misc/gravity\"><span style=\" text-decoration: underline; color:#0000ff;\">Click here</span></a> to see which rooms are affected.</p><p>Enabling this option adds these alternatives to logic. Be careful that this means collecting Gravity Suit can cause items to be permanently inacessible!</p></body></html>", None))
        self.trick_level_help_label.setText(QCoreApplication.translate("PresetTrickLevel", u"<html><head/><body><p>If you want to tweak the knowledge or skill needed expected in a game, you can configure the level used for each of the tweaks listed below by moving the slider to the apropriate level.</p><p>Press the ? button to see which rooms use that trick on the selected level.</p></body></html>", None))
    # retranslateUi




from randovania.gui.generated.preset_logic_damage_ui import Ui_PresetLogicDamage
from randovania.gui.lib import common_qt_lib
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.damage_strictness import LayoutDamageStrictness
from randovania.layout.preset import Preset


class PresetLogicDamage(PresetTab, Ui_PresetLogicDamage):

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        self.damage_strictness_combo.setItemData(0, LayoutDamageStrictness.STRICT)
        self.damage_strictness_combo.setItemData(1, LayoutDamageStrictness.MEDIUM)
        self.damage_strictness_combo.setItemData(2, LayoutDamageStrictness.LENIENT)
        self.damage_strictness_combo.currentIndexChanged.connect(self._on_update_damage_strictness)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def on_preset_changed(self, preset: Preset):
        common_qt_lib.set_combo_with_value(self.damage_strictness_combo, preset.configuration.damage_strictness)

    def _on_update_damage_strictness(self, new_index: int):
        with self._editor as editor:
            editor.layout_configuration_damage_strictness = self.damage_strictness_combo.currentData()



import dataclasses
import dataclasses
import logging
import struct
from typing import List, Tuple, Set, Optional

from randovania.dol_patching import assembler
from randovania.game_connection.connection_base import InventoryItem, Inventory
from randovania.game_connection.connector.remote_connector import RemoteConnector, RemotePatch
from randovania.game_connection.executor.memory_operation import MemoryOperationException, MemoryOperation, \
    MemoryOperationExecutor
from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import (
    CurrentResources, add_resource_gain_to_current_resources
)
from randovania.game_description.world.world import World
from randovania.games.game import RandovaniaGame
from randovania.games.prime import (all_prime_dol_patches)


@dataclasses.dataclass(frozen=True)
class DolRemotePatch(RemotePatch):
    instructions: List[assembler.BaseInstruction]


class PrimeRemoteConnector(RemoteConnector):
    version: all_prime_dol_patches.BasePrimeDolVersion
    game: GameDescription
    _last_message_size: int = 0

    def __init__(self, version: all_prime_dol_patches.BasePrimeDolVersion):
        super().__init__()
        self.logger = logging.getLogger(type(self).__name__)

        self.version = version
        self.game = default_database.game_description_for(version.game)

    @property
    def game_enum(self) -> RandovaniaGame:
        return self.version.game

    async def is_this_version(self, executor: MemoryOperationExecutor) -> bool:
        """Returns True if the accessible memory matches the version of this connector."""
        operation = MemoryOperation(self.version.build_string_address, read_byte_count=len(self.version.build_string))
        build_string = await executor.perform_single_memory_operation(operation)
        return build_string == self.version.build_string

    def _asset_id_format(self):
        """struct.unpack format string for decoding an asset id"""
        raise NotImplementedError()

    def _current_status_world(self, world_asset_id: Optional[bytes], vtable_bytes: Optional[bytes]) -> Optional[World]:
        """
        Helper for `current_game_status`. Calculates the current World based on raw world_asset_id and vtable pointer.
        :param world_asset_id: Bytes for the current world asset id. Might be None.
        :param vtable_bytes: Bytes for a CPlayer vtable pointer. Might be None.
        :return:
        """

        if world_asset_id is None:
            return None

        if self.version.cplayer_vtable is not None:
            if vtable_bytes is None:
                return None

            player_vtable = struct.unpack(">I", vtable_bytes)[0]
            if player_vtable != self.version.cplayer_vtable:
                return None

        asset_id = struct.unpack(self._asset_id_format(), world_asset_id)[0]
        try:
            new_world = self.game.world_list.world_by_asset_id(asset_id)
        except KeyError:
            return None

        return new_world

    async def current_game_status(self, executor: MemoryOperationExecutor):
        raise NotImplementedError()

    async def _memory_op_for_items(self, executor: MemoryOperationExecutor, items: List[ItemResourceInfo],
                                   ) -> List[MemoryOperation]:
        raise NotImplementedError()

    async def get_inventory(self, executor: MemoryOperationExecutor) -> Inventory:
        """Fetches the inventory represented by the given game memory."""

        memory_ops = await self._memory_op_for_items(executor, [
            item
            for item in self.game.resource_database.item
            if item.index < 1000
        ])
        ops_result = await executor.perform_memory_operations(memory_ops)

        inventory = {}
        for item, memory_op in zip(self.game.resource_database.item, memory_ops):
            inv = InventoryItem(*struct.unpack(">II", ops_result[memory_op]))
            if (inv.amount > inv.capacity or inv.capacity > item.max_capacity) and (
                    item != self.game.resource_database.multiworld_magic_item):
                raise MemoryOperationException(f"Received {inv} for {item.long_name}, which is an invalid state.")
            inventory[item] = inv

        return inventory

    async def known_collected_locations(self, executor: MemoryOperationExecutor,
                                        ) -> Tuple[Set[PickupIndex], List[DolRemotePatch]]:
        """Fetches pickup indices that have been collected.
        The list may return less than all collected locations, depending on implementation details.
        This function also returns a list of remote patches that must be performed via `execute_remote_patches`.
        """
        multiworld_magic_item = self.game.resource_database.multiworld_magic_item
        if multiworld_magic_item is None:
            return set(), []

        memory_ops = await self._memory_op_for_items(executor, [multiworld_magic_item])
        op_result = await executor.perform_single_memory_operation(*memory_ops)

        magic_inv = InventoryItem(*struct.unpack(">II", op_result))
        if magic_inv.amount > 0:
            self.logger.info(f"magic item was at {magic_inv.amount}/{magic_inv.capacity}")
            locations = {PickupIndex(magic_inv.amount - 1)}
            patches = [DolRemotePatch([], all_prime_dol_patches.adjust_item_amount_and_capacity_patch(
                self.version.powerup_functions,
                self.game.game,
                multiworld_magic_item.index,
                -magic_inv.amount,
            ))]
            return locations, patches
        else:
            return set(), []

    async def find_missing_remote_pickups(self, executor: MemoryOperationExecutor, inventory: Inventory,
                                          remote_pickups: Tuple[Tuple[str, PickupEntry], ...],
                                          in_cooldown: bool,
                                          ) -> Tuple[List[RemotePatch], bool]:
        """
        Determines if any of the remote_pickups needs to be written to executor.
        :param executor:
        :param inventory: The player's inventory, as given by `get_inventory`.
        :param remote_pickups: Ordered list of pickups sent from other players, with the name of the player.
        :param in_cooldown: If sending new pickups is on cooldown.
        :return: List of patches to give one missing pickup. A bool indicating that a message will be displayed.
        """
        multiworld_magic_item = self.game.resource_database.multiworld_magic_item
        magic_inv = inventory.get(multiworld_magic_item)
        if magic_inv is None or magic_inv.amount > 0 or magic_inv.capacity >= len(remote_pickups) or in_cooldown:
            return [], False

        provider_name, pickup = remote_pickups[magic_inv.capacity]
        item_patches, message = await self._patches_for_pickup(provider_name, pickup, inventory)
        self.logger.info(f"{len(remote_pickups)} permanent pickups, magic {magic_inv.capacity}. "
                         f"Next pickup: {message}")

        patches = [DolRemotePatch([], item_patch) for item_patch in item_patches]
        patches.append(DolRemotePatch([], all_prime_dol_patches.increment_item_capacity_patch(
            self.version.powerup_functions,
            self.game.game,
            multiworld_magic_item.index,
        )))
        patches.append(self._dol_patch_for_hud_message(message))

        return patches, True

    async def execute_remote_patches(self, executor: MemoryOperationExecutor, patches: List[DolRemotePatch]) -> None:
        """
        Executes a given set of patches on the given memory operator. Should only be called if the bool returned by
        `current_game_status` is False, but validation of this fact is implementation-dependant.
        :param executor:
        :param patches: List of patches to execute
        :return:
        """
        memory_operations = []
        for patch in patches:
            memory_operations.extend(patch.memory_operations)

        patch_address, patch_bytes = all_prime_dol_patches.create_remote_execution_body(
            self.version.string_display,
            [instruction
             for patch in patches
             for instruction in patch.instructions],
        )
        memory_operations.extend([
            MemoryOperation(patch_address, write_bytes=patch_bytes),
            MemoryOperation(self.version.cstate_manager_global + 0x2, write_bytes=b"\x01"),
        ])
        self.logger.debug(f"Performing {len(memory_operations)} ops with {len(patches)} patches")
        await executor.perform_memory_operations(memory_operations)

    def _resources_to_give_for_pickup(self, pickup: PickupEntry, inventory: Inventory) -> Tuple[str, CurrentResources]:
        inventory_resources: CurrentResources = {
            item: inv_item.capacity
            for item, inv_item in inventory.items()
        }
        conditional = pickup.conditional_for_resources(inventory_resources)
        if conditional.name is not None:
            item_name = conditional.name
        else:
            item_name = pickup.name

        resources_to_give = {}

        if pickup.respects_lock and not pickup.unlocks_resource and (
                pickup.resource_lock is not None and inventory_resources.get(pickup.resource_lock.locked_by, 0) == 0):
            pickup_resources = list(pickup.resource_lock.convert_gain(conditional.resources))
            item_name = f"Locked {item_name}"
        else:
            pickup_resources = conditional.resources

        add_resource_gain_to_current_resources(pickup_resources, inventory_resources)
        add_resource_gain_to_current_resources(pickup_resources, resources_to_give)
        add_resource_gain_to_current_resources(pickup.conversion_resource_gain(inventory_resources),
                                               resources_to_give)

        # Ignore item% for received items
        if self.game.resource_database.item_percentage is not None:
            resources_to_give.pop(self.game.resource_database.item_percentage, None)

        return item_name, resources_to_give

    async def _patches_for_pickup(self, provider_name: str, pickup: PickupEntry, inventory: Inventory
                                  ) -> Tuple[List[List[assembler.BaseInstruction]], str]:
        raise NotImplementedError()

    def _write_string_to_game_buffer(self, message: str) -> MemoryOperation:
        overhead_size = 6  # 2 bytes for an extra char to differentiate sizes
        encoded_message = message.encode("utf-16_be")[:self.version.string_display.max_message_size - overhead_size]

        # The game doesn't handle very well a string at the same address with same size being
        # displayed multiple times
        if len(encoded_message) == self._last_message_size:
            encoded_message += b'\x00 '
        self._last_message_size = len(encoded_message)

        # Add the null terminator
        encoded_message += b"\x00\x00"
        if len(encoded_message) & 3:
            # Ensure the size is a multiple of 4
            num_to_align = (len(encoded_message) | 3) - len(encoded_message) + 1
            encoded_message += b"\x00" * num_to_align

        return MemoryOperation(self.version.string_display.message_receiver_string_ref,
                               write_bytes=encoded_message)

    def _dol_patch_for_hud_message(self, message: str) -> DolRemotePatch:
        return DolRemotePatch(
            [self._write_string_to_game_buffer(message)],
            all_prime_dol_patches.call_display_hud_patch(self.version.string_display),
        )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'login_prompt_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LoginPromptDialog(object):
    def setupUi(self, LoginPromptDialog):
        if not LoginPromptDialog.objectName():
            LoginPromptDialog.setObjectName(u"LoginPromptDialog")
        LoginPromptDialog.resize(331, 233)
        self.layout = QVBoxLayout(LoginPromptDialog)
        self.layout.setSpacing(6)
        self.layout.setContentsMargins(11, 11, 11, 11)
        self.layout.setObjectName(u"layout")
        self.title_label = QLabel(LoginPromptDialog)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setWordWrap(True)

        self.layout.addWidget(self.title_label)

        self.discord_button = QPushButton(LoginPromptDialog)
        self.discord_button.setObjectName(u"discord_button")

        self.layout.addWidget(self.discord_button)

        self.guest_button = QPushButton(LoginPromptDialog)
        self.guest_button.setObjectName(u"guest_button")

        self.layout.addWidget(self.guest_button)

        self.connection_status_label = QLabel(LoginPromptDialog)
        self.connection_status_label.setObjectName(u"connection_status_label")
        self.connection_status_label.setAlignment(Qt.AlignCenter)

        self.layout.addWidget(self.connection_status_label)

        self.spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.spacer)

        self.privacy_policy_label = QLabel(LoginPromptDialog)
        self.privacy_policy_label.setObjectName(u"privacy_policy_label")
        self.privacy_policy_label.setAlignment(Qt.AlignBottom|Qt.AlignHCenter)
        self.privacy_policy_label.setWordWrap(True)
        self.privacy_policy_label.setOpenExternalLinks(True)

        self.layout.addWidget(self.privacy_policy_label)

        self.button_box = QDialogButtonBox(LoginPromptDialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Ok|QDialogButtonBox.Reset)

        self.layout.addWidget(self.button_box)


        self.retranslateUi(LoginPromptDialog)

        QMetaObject.connectSlotsByName(LoginPromptDialog)
    # setupUi

    def retranslateUi(self, LoginPromptDialog):
        LoginPromptDialog.setWindowTitle(QCoreApplication.translate("LoginPromptDialog", u"Login Window", None))
        self.title_label.setText(QCoreApplication.translate("LoginPromptDialog", u"<html><head/><body><p>Usage of Randovania's online services depends on an user account.</p><p>Currently, only login via Discord is supported.</p></body></html>", None))
        self.discord_button.setText(QCoreApplication.translate("LoginPromptDialog", u"Login with Discord", None))
        self.guest_button.setText(QCoreApplication.translate("LoginPromptDialog", u"Login as Guest", None))
        self.connection_status_label.setText("")
        self.privacy_policy_label.setText(QCoreApplication.translate("LoginPromptDialog", u"<html><head/><body><p>By creating an account, you are agreeing to Randovania's <a href=\"https://github.com/randovania/randovania/blob/master/PRIVACY_POLICY.md\"><span style=\" text-decoration: underline; color:#0000ff;\">Privacy Policy</span></a>.</p></body></html>", None))
    # retranslateUi




from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class LayoutLogicalResourceAction(BitPackEnum, Enum):
    RANDOMLY = "randomly"
    LAST_RESORT = "last_resort"
    NEVER = "never"

    @property
    def long_name(self) -> str:
        return PRETTY_NAME[self]


PRETTY_NAME = {
    LayoutLogicalResourceAction.RANDOMLY: "Randomly",
    LayoutLogicalResourceAction.LAST_RESORT: "Last resort",
    LayoutLogicalResourceAction.NEVER: "Never",
}



import dataclasses
import datetime
from typing import List, Optional

import aiohttp
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QPushButton, QDialogButtonBox, QDialog, QTableWidgetItem, QCheckBox
from qasync import asyncSlot

from randovania.games.game import RandovaniaGame
from randovania.gui.generated.racetime_browser_dialog_ui import Ui_RacetimeBrowserDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.qt_network_client import handle_network_errors
from randovania.layout.permalink import Permalink


@dataclasses.dataclass(frozen=True)
class RaceEntry:
    name: str
    status: str
    verbose_status: str
    entrants: int
    goal: str
    opened_at: datetime.datetime
    info: str
    game: RandovaniaGame

    @classmethod
    def from_json(cls, data, game) -> "RaceEntry":
        return RaceEntry(
            name=data["name"],
            status=data["status"]["value"],
            verbose_status=data["status"]["verbose_value"],
            entrants=data["entrants_count"],
            goal=data["goal"]["name"],
            opened_at=datetime.datetime.fromisoformat(data["opened_at"].replace("Z", "+00:00")),
            info=data["info"],
            game=game,
        )

_SUPPORTED_GAME_URLS = {
    RandovaniaGame.METROID_PRIME: "https://racetime.gg/mpr/data",
    RandovaniaGame.METROID_PRIME_ECHOES: "https://racetime.gg/mp2r/data",
}
_TEST_RESPONSE = {
    "name": "Metroid Prime 2: Echoes Randomizer",
    "short_name": "MP2R",
    "slug": "mp2r",
    "url": "/mp2r",
    "data_url": "/mp2r/data",
    "image": None,
    "info": "Download the Metroid Prime 2: Echoes randomizer <a href=\"https://github.com/randovania/randovania/releases\">here</a>.\r\n<p><p>\r\n<b><u>Category Rules</b></u>\r\n<p><p>\r\nTiming starts when confirming the game's difficulty on the main menu (or selecting a new file, if you are not able to select a difficulty). \r\n<p><p>\r\n<b>Beat the game</b>: Timing stops once you have activated a cutscene, elevator, portal, or teleport that will lead to the credits or Mission Final screen without any further game input.\r\n <p>\r\n<b>Collect all Sky Temple Keys</b>: Timing stops once you have at least one of each available Sky Temple Key (Sky Temple Key 1, 2, 3, etc.) currently collected. \r\n<p>\r\n<b><u>Miscellaneous</b></u>\r\n<p><p>\r\nThe only permitted emulator is Dolphin, running on version 5.0 or later. Running the ISO on console via Nintendont is allowed.\r\n<p>\r\nISO inspection/ROM inspection/memory inspection/software reverse-engineering that determines item locations or elevator destinations of the race seed prior to or during the race is <b>banned</b>.\r\n<p>\r\nTurbo functions, increasing the disk read speed, save states, fast forward and disabling fog are all <b>banned</b>. However, any of these functions may be allowed if all racers permit it, or a category rule permits it.",
    "streaming_required": False,
    "owners": [
        {
            "id": "q5rNGD3DYAo9blOy",
            "full_name": "gollop#9668",
            "name": "gollop",
            "discriminator": "9668",
            "url": "/user/q5rNGD3DYAo9blOy",
            "avatar": None,
            "pronouns": "they/them",
            "flair": "moderator",
            "twitch_name": "golloppp",
            "twitch_display_name": "golloppp",
            "twitch_channel": "https://www.twitch.tv/golloppp",
            "can_moderate": True
        }
    ],
    "moderators": [],
    "current_races": [
        {
            "name": "mp2r/proud-stadium-7340",
            "status": {
                "value": "in_progress",
                "verbose_value": "In progress",
                "help_text": "Race is in progress"
            },
            "url": "/mp2r/proud-stadium-7340",
            "data_url": "/mp2r/proud-stadium-7340/data",
            "goal": {
                "name": "Beat the game",
                "custom": False
            },
            "info": "zasj0oGRVOUaUwbx6LFMgLrjmF0kH2aEFX9OMr1tu7xcW8az || Seed Hash: Command Caretaker Alcove (67J3ZJPT)",
            "entrants_count": 3,
            "entrants_count_inactive": 0,
            "opened_at": "2020-10-03T18:44:17.865Z",
            "started_at": "2020-10-03T18:45:17.865Z",
            "time_limit": "P1DT00H00M00S"
        },
        {
            "name": "mp2r/proud-arena-4584",
            "status": {
                "value": "open",
                "verbose_value": "Open",
                "help_text": "Anyone may join this race"
            },
            "url": "/mp2r/proud-arena-4584",
            "data_url": "/mp2r/proud-arena-4584/data",
            "goal": {
                "name": "Beat the game",
                "custom": False
            },
            "info": "yevhOB287u6SOly2HFReAapCZprQOvB9fVaU5-EAvS08qoQi  || Seed Hash: Sandigger Module Boom (GN73RSFH)",
            "entrants_count": 5,
            "entrants_count_inactive": 0,
            "opened_at": "2020-10-03T18:46:17.865Z",
            "started_at": None,
            "time_limit": "P1DT00H00M00S"
        }
    ]
}


async def _query_server(race_url) -> dict:
    async with aiohttp.ClientSession() as session:
        async with session.get(race_url) as response:
            response.raise_for_status()
            return await response.json()


class RacetimeBrowserDialog(QDialog, Ui_RacetimeBrowserDialog):
    races: List[RaceEntry]
    permalink: Optional[Permalink] = None

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.refresh_button = QPushButton("Refresh")
        self.button_box.addButton(self.refresh_button, QDialogButtonBox.ResetRole)
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.button_box.button(QDialogButtonBox.Ok).setText("Import")

        self.button_box.accepted.connect(self.attempt_join)
        self.button_box.rejected.connect(self.reject)
        self.refresh_button.clicked.connect(self.refresh)

        self._status_checks = (
            (self.status_open_check, "open"),
            (self.status_invitational_check, "invitational"),
            (self.status_pending_check, "pending"),
            (self.status_inprogress_check, "in_progress"),
            (self.status_finished_check, "finished"),
            (self.status_cancelled_check, "cancelled"),
        )
        for check, _ in self._status_checks:
            check.stateChanged.connect(self.update_list)
        self.filter_name_edit.textEdited.connect(self.update_list)

        self._game_checks = set()
        for game in _SUPPORTED_GAME_URLS:
            game_checkbox = QCheckBox(game.long_name)
            game_checkbox.setChecked(True)
            self.game_check_layout.addWidget(game_checkbox)
            game_checkbox.stateChanged.connect(self.update_list)
            self._game_checks.add((game_checkbox, game))

        self.table_widget.itemSelectionChanged.connect(self.on_selection_changed)
        self.table_widget.itemDoubleClicked.connect(self.on_double_click)

    @asyncSlot()
    @handle_network_errors
    async def refresh(self) -> bool:
        self.refresh_button.setEnabled(False)
        self.races = []
        try:
            for game in _SUPPORTED_GAME_URLS:
                race_url = _SUPPORTED_GAME_URLS[game]
                try:
                    raw_races = await _query_server(race_url)
                except aiohttp.ClientError as e:
                    await async_dialog.warning(self, "Connection error",
                                               f"Unable to retrieve races from `{race_url}`: {e}")
                    return False

                self.races.extend([RaceEntry.from_json(item, game) for item in raw_races["current_races"]])
            self.update_list()
            return True
        finally:
            self.refresh_button.setEnabled(True)

    def on_selection_changed(self):
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(len(self.table_widget.selectedItems()) > 0)

    @property
    def selected_race(self):
        return self.table_widget.selectedItems()[0].data(Qt.UserRole)

    @asyncSlot(QTableWidgetItem)
    async def on_double_click(self, item: QTableWidgetItem):
        await self.attempt_join()

    @asyncSlot()
    @handle_network_errors
    async def attempt_join(self):
        if not self.races:
            return

        race = self.selected_race

        permalink = None

        for word in race.info.split(" "):
            try:
                permalink = Permalink.from_str(word)
            except ValueError:
                continue

        if permalink is None:
            await async_dialog.warning(self, "Missing permalink",
                                       "Unable to get a valid Permalink from this race's info.")
        else:
            self.permalink = permalink
            return self.accept()

    def update_list(self):
        self.table_widget.clear()
        self.table_widget.setHorizontalHeaderLabels(["Name", "Game", "Status", "Entrants",
                                                     "Goal", "Info", "Opened At"])

        name_filter = self.filter_name_edit.text().strip()

        displayed_status = set()
        for (check, status) in self._status_checks:
            if check.isChecked():
                displayed_status.add(status)

        displayed_games = set()
        for (check, game) in self._game_checks:
            if check.isChecked():
                displayed_games.add(game)

        visible_races = [
            race
            for race in self.races
            if (race.status in displayed_status
                and name_filter in race.name
                and race.game in displayed_games)
        ]

        self.table_widget.setRowCount(len(visible_races))
        for i, session in enumerate(visible_races):
            name = QTableWidgetItem(session.name)
            status = QTableWidgetItem(session.verbose_status)
            entrants = QTableWidgetItem(str(session.entrants))
            goal = QTableWidgetItem(session.goal)
            info = QTableWidgetItem(session.info)
            opened_at = QTableWidgetItem(session.opened_at.astimezone().strftime("%c"))
            game_name = QTableWidgetItem(session.game.short_name)

            name.setData(Qt.UserRole, session)
            self.table_widget.setItem(i, 0, name)
            self.table_widget.setItem(i, 1, game_name)
            self.table_widget.setItem(i, 2, status)
            self.table_widget.setItem(i, 3, entrants)
            self.table_widget.setItem(i, 4, goal)
            self.table_widget.setItem(i, 5, info)
            self.table_widget.setItem(i, 6, opened_at)

        self.status_label.setText(f"{len(self.races)} races total, {len(visible_races)} displayed.")



import asyncio
import platform
import re
from asyncio import StreamWriter, StreamReader, IncompleteReadError
from pathlib import Path
from typing import Callable, List, Optional

IO_LOOP: Optional[asyncio.AbstractEventLoop] = None


def is_windows() -> bool:
    return platform.system() == "Windows"


async def _write_data(stream: StreamWriter, data: str):
    stream.write(data.encode("UTF-8"))
    stream.close()


async def _read_data(stream: StreamReader, read_callback: Callable[[str], None]):
    while True:
        try:
            line = await stream.readuntil(b"\r")
        except IncompleteReadError as incomplete:
            line = incomplete.partial
        if line:
            try:
                decoded = line.decode()
            except UnicodeDecodeError:
                decoded = line.decode("latin1")
            for x in re.split(r"[\r\n]", decoded.strip()):
                if x:
                    read_callback(x)
        else:
            break


async def _process_command_async(args: List[str], input_data: str, read_callback: Callable[[str], None]):
    process = await asyncio.create_subprocess_exec(*args,
                                                   stdin=asyncio.subprocess.PIPE,
                                                   stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)

    await asyncio.wait([
        _write_data(process.stdin, input_data),
        _read_data(process.stdout, read_callback),
    ])
    await process.wait()


def process_command(args: List[str], input_data: str, read_callback: Callable[[str], None],
                    add_mono_if_needed: bool = True):
    if not Path(args[0]).is_file():
        raise FileNotFoundError("{} not found".format(args[0]))

    if add_mono_if_needed and not is_windows():
        args = ["mono", *args]
    work = _process_command_async(args, input_data, read_callback)

    if IO_LOOP is None:
        if is_windows():
            asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
        asyncio.run(work)
    else:
        asyncio.run_coroutine_threadsafe(work, IO_LOOP).result()



import dataclasses
from typing import Iterator

from randovania.bitpacking.bitpacking import BitPackDataclass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.bitpacking.type_enforcement import DataclassPostInitTypeCheck


@dataclasses.dataclass(frozen=True)
class BeamAmmoConfiguration(BitPackDataclass, JsonDataclass, DataclassPostInitTypeCheck):
    item_index: int = dataclasses.field(metadata={"min": 0, "max": 108})
    ammo_a: int = dataclasses.field(metadata={"min": -1, "max": 108})
    ammo_b: int = dataclasses.field(metadata={"min": -1, "max": 108})
    uncharged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 1})
    charged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 5})
    combo_missile_cost: int = dataclasses.field(metadata={"min": 1, "max": 250, "if_different": 5})
    combo_ammo_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 30})


@dataclasses.dataclass(frozen=True)
class BeamConfiguration(BitPackDataclass, JsonDataclass):
    power: BeamAmmoConfiguration
    dark: BeamAmmoConfiguration
    light: BeamAmmoConfiguration
    annihilator: BeamAmmoConfiguration

    @property
    def all_beams(self) -> Iterator[BeamAmmoConfiguration]:
        yield self.power
        yield self.dark
        yield self.light
        yield self.annihilator



import dataclasses
from unittest.mock import MagicMock

import pytest

from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.prime2.echoes_configuration import LayoutSkyTempleKeyMode


@pytest.fixture(name="editor")
def _editor() -> PresetEditor:
    return PresetEditor(MagicMock())


_sample_layout_configurations = [
    {
        "sky_temple_keys": LayoutSkyTempleKeyMode.default(),
        "menu_mod": False,
    }
]


@pytest.fixture(params=_sample_layout_configurations, name="initial_layout_configuration_params")
def _initial_layout_configuration_params(request) -> dict:
    return request.param


@pytest.mark.parametrize("menu_mod", [False, True])
def test_edit_menu_mod(editor: PresetEditor,
                       initial_layout_configuration_params: dict,
                       default_layout_configuration,
                       menu_mod):
    # Setup
    editor._configuration = dataclasses.replace(default_layout_configuration,
                                                **initial_layout_configuration_params)
    editor._nested_autosave_level = 1

    # Run
    initial_layout_configuration_params["menu_mod"] = menu_mod
    editor.set_configuration_field("menu_mod", menu_mod)

    # Assert
    assert editor.configuration == dataclasses.replace(default_layout_configuration,
                                                       **initial_layout_configuration_params)



import json
from pathlib import Path

import pytest
from mock import MagicMock, AsyncMock, call

from randovania.game_connection.game_connection import GameConnection
from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.gui import multiworld_client
from randovania.gui.multiworld_client import MultiworldClient, Data


@pytest.fixture(name="client")
def _client(skip_qtbot):
    network_client = MagicMock()
    game_connection = MagicMock(spec=GameConnection)
    game_connection.lock_identifier = None
    return MultiworldClient(network_client, game_connection)


@pytest.mark.asyncio
async def test_start(client, tmpdir):
    game_connection = client.game_connection

    client.network_client.game_session_request_pickups = AsyncMock(return_value=[])
    client.network_client.game_session_request_update = AsyncMock()
    client.network_client.session_self_update = AsyncMock()

    # Run
    await client.start(Path(tmpdir).joinpath("missing_file.json"))

    # Assert
    game_connection.set_location_collected_listener.assert_called_once_with(client.on_location_collected)
    client.network_client.GameSessionPickupsUpdated.connect.assert_called_once_with(client.on_network_game_updated)


@pytest.mark.asyncio
async def test_stop(client: MultiworldClient):
    # Run
    await client.stop()

    # Assert
    client.game_connection.set_location_collected_listener.assert_called_once_with(None)
    client.network_client.GameSessionPickupsUpdated.disconnect.assert_called_once_with(client.on_network_game_updated)
    client.game_connection.set_expected_game.assert_called_once_with(None)
    client.game_connection.set_permanent_pickups.assert_called_once_with(())


@pytest.mark.parametrize("wrong_game", [False, True])
@pytest.mark.parametrize("exists", [False, True])
@pytest.mark.asyncio
async def test_on_location_collected(client: MultiworldClient, tmpdir, exists, wrong_game):
    client._data = Data(Path(tmpdir).joinpath("data.json"))
    client._data.collected_locations = {10, 15} if exists else {10}
    client.start_notify_collect_locations_task = MagicMock()

    if wrong_game and not exists:
        expected_locations = {10}
    else:
        expected_locations = {10, 15}

    if not wrong_game:
        client._expected_game = RandovaniaGame.METROID_PRIME_ECHOES

    # Run
    await client.on_location_collected(RandovaniaGame.METROID_PRIME_ECHOES, PickupIndex(15))

    # Assert
    assert client._data.collected_locations == expected_locations

    if exists or wrong_game:
        client.start_notify_collect_locations_task.assert_not_called()
    else:
        client.start_notify_collect_locations_task.assert_called_once_with()


@pytest.mark.asyncio
async def test_on_game_updated(client, tmpdir):
    client._data = Data(Path(tmpdir).joinpath("data.json"))
    pickups = MagicMock()

    # Run
    await client.on_network_game_updated(pickups)

    # Assert
    client.game_connection.set_expected_game.assert_called_once_with(pickups.game)
    client.game_connection.set_permanent_pickups.assert_called_once_with(pickups.pickups)


@pytest.mark.asyncio
async def test_notify_collect_locations(client, tmpdir):
    data_path = Path(tmpdir).joinpath("data.json")
    network_client = client.network_client
    network_client.game_session_collect_locations = AsyncMock(side_effect=[
        RuntimeError("connection issue!"),
        None,
    ])

    data_path.write_text(json.dumps({
        "collected_locations": [10, 15],
        "uploaded_locations": [15],
        "latest_message_displayed": 0,
    }))
    client._data = Data(data_path)

    # Run
    await client._notify_collect_locations()

    # Assert
    network_client.game_session_collect_locations.assert_has_awaits([call((10,)), call((10,))])
    assert set(json.loads(data_path.read_text())["uploaded_locations"]) == {10, 15}


@pytest.mark.asyncio
async def test_lock_file_on_init(skip_qtbot, tmpdir):
    # Setup
    network_client = MagicMock()
    network_client.game_session_request_update = AsyncMock()
    network_client.session_self_update = AsyncMock()
    game_connection = MagicMock(spec=GameConnection)
    game_connection.lock_identifier = str(tmpdir.join("my-lock"))

    # Run
    client = MultiworldClient(network_client, game_connection)
    assert tmpdir.join("my-lock.pid").exists()

    await client.start(Path(tmpdir).joinpath("data.json"))
    await client.stop()
    assert not tmpdir.join("my-lock.pid").exists()



import contextlib
from typing import Optional, List

import pytest
from mock import AsyncMock, MagicMock, call

from randovania.game_connection.connection_backend import ConnectionBackend
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_connection.connector.echoes_remote_connector import EchoesRemoteConnector
from randovania.game_connection.executor.memory_operation import MemoryOperationException, MemoryOperation, \
    MemoryOperationExecutor
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.prime import echoes_dol_versions


@pytest.fixture(name="backend")
def dolphin_backend():
    backend = ConnectionBackend(MagicMock(spec=MemoryOperationExecutor))
    return backend


@pytest.mark.asyncio
async def test_identify_game_ntsc(backend: ConnectionBackend):
    # Setup
    def side_effect(ops: List[MemoryOperation]):
        if len(ops) > 1:
            return {
                op: b"!#$M"
                for op in ops
                if op.address == 0x803ac3b0
            }
        return {}

    build_info = b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32"
    backend.executor.perform_memory_operations.side_effect = side_effect
    backend.executor.perform_single_memory_operation.return_value = build_info

    # Run
    connector = await backend._identify_game()

    # Assert
    assert isinstance(connector, EchoesRemoteConnector)
    assert connector.version is echoes_dol_versions.ALL_VERSIONS[0]


@pytest.mark.asyncio
async def test_identify_game_error(backend: ConnectionBackend):
    # Setup
    backend.executor.perform_memory_operations.side_effect = RuntimeError("not connected")

    # Run
    assert not await backend._identify_game()


@pytest.mark.asyncio
async def test_identify_game_already_known(backend: ConnectionBackend):
    # Setup
    backend.connector = True
    backend.executor.perform_memory_operations.side_effect = RuntimeError("not connected")
    backend.executor.perform_single_memory_operation.return_value = RuntimeError("not connected")

    # Run
    connector = await backend._identify_game()

    # Assert
    assert connector is None


@pytest.mark.parametrize("failure_at", [None, 1, 2])
@pytest.mark.parametrize("depth", [0, 1, 2])
@pytest.mark.asyncio
async def test_interact_with_game(backend: ConnectionBackend, depth: int, failure_at: Optional[int]):
    # Setup
    backend.message_cooldown = 2
    backend.connector = AsyncMock()

    backend.connector.current_game_status.return_value = (
        depth <= 1,  # has pending op
        MagicMock() if depth > 0 else None,  # world
    )

    expectation = contextlib.nullcontext()
    if failure_at == 1:
        backend.connector.get_inventory.side_effect = MemoryOperationException("error at _get_inventory")
        expectation = pytest.raises(MemoryOperationException, match="error at _get_inventory")

    backend._multiworld_interaction = AsyncMock()
    if failure_at == 2:
        backend._multiworld_interaction.side_effect = MemoryOperationException("error at _check_for_collected_index")
        expectation = pytest.raises(MemoryOperationException, match="error at _check_for_collected_index")

    expected_depth = min(depth, failure_at) if failure_at is not None else depth
    if (failure_at or 999) > depth:
        expectation = contextlib.nullcontext()

    # Run
    with expectation:
        await backend._interact_with_game(1)

    # Assert
    assert backend.message_cooldown == (2 if expected_depth < 2 else 1)
    backend.connector.current_game_status.assert_awaited_once_with(backend.executor)

    if expected_depth > 0:
        backend.connector.get_inventory.assert_awaited_once_with(backend.executor)
    else:
        backend.connector.get_inventory.assert_not_awaited()

    if expected_depth > 1:
        backend._multiworld_interaction.assert_awaited_once_with()
    else:
        backend._multiworld_interaction.assert_not_awaited()

    if 0 < depth:
        assert backend._world is not None
    else:
        assert backend._world is None


def test_current_status_disconnected(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = False
    assert backend.current_status == GameConnectionStatus.Disconnected


def test_current_status_unknown_game(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = True
    assert backend.current_status == GameConnectionStatus.UnknownGame


def test_current_status_wrong_game(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = True
    backend.connector = MagicMock()
    backend.connector.game_enum = RandovaniaGame.METROID_PRIME
    backend.set_expected_game(RandovaniaGame.METROID_PRIME_ECHOES)
    assert backend.current_status == GameConnectionStatus.WrongGame


def test_current_status_not_in_game(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = True
    backend.connector = True
    assert backend.current_status == GameConnectionStatus.TitleScreen


def test_current_status_tracker_only(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = True
    backend.connector = True
    backend._world = True
    assert backend.current_status == GameConnectionStatus.TrackerOnly


def test_current_status_in_game(backend: ConnectionBackend):
    backend.executor.is_connected.return_value = True
    backend.connector = True
    backend._world = True
    backend.checking_for_collected_index = True
    assert backend.current_status == GameConnectionStatus.InGame


@pytest.mark.parametrize("depth", [0, 1, 2])
@pytest.mark.asyncio
async def test_multiworld_interaction(backend: ConnectionBackend, depth: int):
    # Setup
    # depth 0: wrong game
    # depth 1: non-empty known_collected_locations with patch
    # depth 2: empty known_collected_locations and empty find_missing_remote_pickups

    game_enum = RandovaniaGame.METROID_PRIME_CORRUPTION
    patches = [MagicMock()]

    location_collected = AsyncMock()
    backend.set_location_collected_listener(location_collected)
    if depth > 0:
        backend.set_expected_game(game_enum)

    connector = AsyncMock()
    connector.game_enum = game_enum
    backend.connector = connector
    backend._inventory = MagicMock()
    backend._permanent_pickups = MagicMock()

    connector.find_missing_remote_pickups.return_value = ([], False)
    if depth == 1:
        connector.known_collected_locations.return_value = ([PickupIndex(2), PickupIndex(5)], patches)
    else:
        connector.known_collected_locations.return_value = ([], [])

    # Run
    await backend._multiworld_interaction()

    # Assert
    connector.known_collected_locations.assert_has_awaits(
        [call(backend.executor)]
        if depth > 0 else []
    )
    if depth == 1:
        location_collected.assert_has_awaits([
            call(game_enum, PickupIndex(2)),
            call(game_enum, PickupIndex(5)),
        ])
        connector.execute_remote_patches(backend.executor, patches)
    else:
        location_collected.assert_not_awaited()
        connector.execute_remote_patches.assert_not_awaited()

    if depth == 2:
        connector.find_missing_remote_pickups.assert_awaited_once_with(
            backend.executor, backend._inventory, backend._permanent_pickups, False
        )
    else:
        connector.find_missing_remote_pickups.assert_not_awaited()


@pytest.mark.parametrize("has_message", [False, True])
@pytest.mark.parametrize("has_cooldown", [False, True])
@pytest.mark.parametrize("has_patches", [False, True])
@pytest.mark.asyncio
async def test_multiworld_interaction_missing_remote_pickups(backend: ConnectionBackend, has_message: bool,
                                                             has_cooldown: bool, has_patches: bool):
    # Setup
    if has_cooldown:
        initial_cooldown = 2.0
    else:
        initial_cooldown = 0.0
    backend.message_cooldown = initial_cooldown

    game_enum = RandovaniaGame.METROID_PRIME_CORRUPTION
    backend.set_expected_game(game_enum)
    patches = [MagicMock()]

    connector = AsyncMock()
    connector.game_enum = game_enum
    backend.connector = connector
    backend._inventory = MagicMock()
    backend._permanent_pickups = MagicMock()

    connector.known_collected_locations.return_value = ([], [])
    connector.find_missing_remote_pickups.return_value = (patches if has_patches else [], has_message)

    # Run
    await backend._multiworld_interaction()

    # Assert
    if has_patches and not (has_cooldown and has_message):
        connector.execute_remote_patches.assert_awaited_once_with(backend.executor, patches)
        if has_message:
            assert backend.message_cooldown == 4.0
        else:
            assert backend.message_cooldown == initial_cooldown
    else:
        connector.execute_remote_patches.assert_not_awaited()
        assert backend.message_cooldown == initial_cooldown


@pytest.mark.parametrize("depth", [0, 1, 2, 3])
@pytest.mark.asyncio
async def test_update(backend: ConnectionBackend, depth: int):
    # Setup
    # depth 0: not enabled
    # depth 1: can't connect
    # depth 2: call identify game
    # depth 3: don't call identify game

    connector = MagicMock()
    backend._enabled = depth > 0
    backend.executor.connect = AsyncMock(return_value=depth > 1)
    backend._expected_game = None
    backend._world = True
    backend._identify_game = AsyncMock(return_value=connector)
    backend._interact_with_game = AsyncMock()
    if depth == 3:
        backend.connector = connector

    # Run
    await backend.update(1)

    # Assert
    backend.executor.connect.assert_has_calls([call()] if depth > 0 else [])

    if depth == 2:
        backend._identify_game.assert_awaited_once_with()
        assert backend.connector == connector
    else:
        backend._identify_game.assert_not_awaited()

    if depth >= 2:
        backend._interact_with_game.assert_awaited_once_with(1)
    else:
        backend._interact_with_game.assert_not_awaited()


@pytest.mark.parametrize("connected_game", [None, RandovaniaGame.METROID_PRIME, RandovaniaGame.METROID_PRIME_ECHOES])
@pytest.mark.parametrize("expected_game", [None, RandovaniaGame.METROID_PRIME_ECHOES])
@pytest.mark.parametrize("interact_fails", [False, True])
@pytest.mark.asyncio
async def test_update_calls_interact_with_game(backend: ConnectionBackend, interact_fails,
                                               expected_game, connected_game):
    # Setup
    backend.connector = None
    backend._world = True
    backend.set_expected_game(expected_game)

    if connected_game is not None:
        connector = AsyncMock()
        connector.game_enum = connected_game
    else:
        connector = None

    backend._interact_with_game = AsyncMock(side_effect=MemoryOperationException("err") if interact_fails else None)
    backend._identify_game = AsyncMock(return_value=connector)

    should_call_interact = connected_game is not None
    if should_call_interact:
        should_call_interact = (expected_game == connected_game) or (expected_game is None)

    # Run
    await backend.update(1)

    # Assert
    backend._identify_game.assert_awaited_once_with()
    if should_call_interact:
        backend._interact_with_game.assert_awaited_once_with(1)
        if interact_fails:
            assert backend._world is None
        else:
            assert backend._world is not None
    else:
        backend._interact_with_game.assert_not_awaited()


def test_get_current_inventory(backend: ConnectionBackend):
    # Setup
    inventory = {"a": 5}
    backend._inventory = inventory

    # Run
    result = backend.get_current_inventory()

    # Assert
    assert result is inventory



from typing import Optional

from PySide2 import QtWidgets
from qasync import asyncSlot

from randovania.gui.dialog.login_prompt_dialog import LoginPromptDialog
from randovania.gui.game_session_window import GameSessionWindow
from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.qt_network_client import handle_network_errors, QtNetworkClient
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.online_game_list_window import GameSessionBrowserDialog
from randovania.interface_common.options import Options
from randovania.interface_common.preset_manager import PresetManager
from randovania.network_client.network_client import ConnectionState


class OnlineInteractions(QtWidgets.QWidget):
    network_client: QtNetworkClient
    game_session_window: Optional[GameSessionWindow] = None
    _login_window: Optional[QtWidgets.QDialog] = None

    def __init__(self, window_manager: WindowManager, preset_manager: PresetManager, network_client: QtNetworkClient,
                 main_window: Ui_MainWindow, options: Options):
        super().__init__(window_manager)

        self.window_manager = window_manager
        self.preset_manager = preset_manager
        self.network_client = network_client
        self.main_window = main_window
        self.options = options

        # Signals
        main_window.browse_sessions_button.clicked.connect(self._browse_for_game_session)
        main_window.host_new_game_button.clicked.connect(self._host_game_session)

        # Menu Bar
        main_window.action_login_window.triggered.connect(self._action_login_window)

    async def _game_session_active(self) -> bool:
        if self.game_session_window is None or self.game_session_window.has_closed:
            return False
        else:
            await async_dialog.message_box(
                self,
                QtWidgets.QMessageBox.Critical,
                "Game Session in progress",
                "There's already a game session window open. Please close it first.",
                QtWidgets.QMessageBox.Ok
            )
            self.game_session_window.activateWindow()
            return True

    async def _ensure_logged_in(self) -> bool:
        network_client = self.network_client
        if network_client.connection_state == ConnectionState.Connected:
            return True

        if network_client.connection_state.is_disconnected:
            message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.NoIcon, "Connecting",
                                                "Connecting to server...", QtWidgets.QMessageBox.Cancel,
                                                self)

            connecting = network_client.connect_to_server()
            message_box.rejected.connect(connecting.cancel)
            message_box.show()
            try:
                await connecting
            finally:
                message_box.close()

        if network_client.current_user is None:
            await async_dialog.execute_dialog(LoginPromptDialog(network_client))

        return network_client.current_user is not None

    @asyncSlot()
    @handle_network_errors
    async def _browse_for_game_session(self):
        if await self._game_session_active():
            return

        if not await self._ensure_logged_in():
            return

        network_client = self.network_client
        browser = GameSessionBrowserDialog(network_client)
        await browser.refresh()
        if await async_dialog.execute_dialog(browser) == browser.Accepted:
            self.game_session_window = await GameSessionWindow.create_and_update(
                network_client, common_qt_lib.get_game_connection(), self.preset_manager,
                self.window_manager, self.options)
            if self.game_session_window is not None:
                self.game_session_window.show()

    @asyncSlot()
    @handle_network_errors
    async def _action_login_window(self):
        if self._login_window is not None:
            return self._login_window.show()

        self._login_window = LoginPromptDialog(self.network_client)
        try:
            await async_dialog.execute_dialog(self._login_window)
        finally:
            self._login_window = None

    @asyncSlot()
    @handle_network_errors
    async def _host_game_session(self):
        if await self._game_session_active():
            return

        if not await self._ensure_logged_in():
            return

        dialog = QtWidgets.QInputDialog(self)
        dialog.setModal(True)
        dialog.setWindowTitle("Enter session name")
        dialog.setLabelText("Select a name for the session:")
        if await async_dialog.execute_dialog(dialog) != dialog.Accepted:
            return

        await self.network_client.create_new_session(dialog.textValue())
        self.game_session_window = await GameSessionWindow.create_and_update(self.network_client,
                                                                             common_qt_lib.get_game_connection(),
                                                                             self.preset_manager,
                                                                             self.window_manager,
                                                                             self.options)
        if self.game_session_window is not None:
            self.game_session_window.show()



import json
import sys
from pathlib import Path
from typing import Optional

from randovania.version import version

CONFIGURATION_FILE_PATH: Optional[Path] = None


def is_frozen() -> bool:
    return getattr(sys, "frozen", False)


def get_data_path() -> Path:
    if is_frozen():
        file_dir = Path(getattr(sys, "_MEIPASS"))
    else:
        file_dir = Path(__file__).parent
    return file_dir.joinpath("data")


def _get_default_configuration_path() -> Path:
    return get_data_path().joinpath("configuration.json")


def get_configuration() -> dict:
    file_path = CONFIGURATION_FILE_PATH
    if file_path is None:
        file_path = _get_default_configuration_path()

    try:
        with file_path.open() as file:
            return json.load(file)
    except FileNotFoundError:
        if CONFIGURATION_FILE_PATH is None:
            return {
                "server_address": "http://127.0.0.1:5000",
                "socketio_path": "/socket.io",
            }
        else:
            raise


def setup_logging(default_level: str, log_to_file: Optional[Path]):
    import logging.config
    import logging.handlers

    handlers = {
        'default': {
            'level': default_level,
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
    }
    if log_to_file is not None:
        handlers['local_app_data'] = {
            'level': 'DEBUG',
            'formatter': 'default',
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': log_to_file,
            'encoding': 'utf-8',
            'backupCount': 10,
        }

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s] [%(levelname)s] [%(name)s] %(funcName)s: %(message)s',
            }
        },
        'handlers': handlers,
        'loggers': {
            'randovania.network_client.network_client': {
                'level': 'DEBUG',
            },
            'randovania.game_connection.connection_backend': {
                'level': 'DEBUG',
            },
            'randovania.gui.multiworld_client': {
                'level': 'DEBUG',
            },
            'NintendontExecutor': {
                'level': 'DEBUG',
            },
            'DolphinExecutor': {
                'level': 'DEBUG',
            },
            'Prime1RemoteConnector': {
                'level': 'DEBUG',
            },
            'EchoesRemoteConnector': {
                'level': 'DEBUG',
            },
            'CorruptionRemoteConnector': {
                'level': 'DEBUG',
            },
            'randovania.gui.qt': {
                'level': 'INFO',
            },
            'qasync': {
                'level': 'INFO',
            },
            # 'socketio.client': {
            #     'level': 'DEBUG',
            # }
        },
        'root': {
            'level': default_level,
            'handlers': list(handlers.keys()),
        },
    })
    logging.info("Logging initialized with level %s for version %s.", default_level, VERSION)


__version__ = version
VERSION = version



from typing import Union, Dict, Type

from randovania.games.game import RandovaniaGame
from randovania.layout.prime1.prime_configuration import PrimeConfiguration
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.layout.prime3.corruption_configuration import CorruptionConfiguration
from randovania.layout.prime3.corruption_cosmetic_patches import CorruptionCosmeticPatches
from randovania.layout.super_metroid.super_metroid_configuration import SuperMetroidConfiguration
from randovania.layout.super_metroid.super_metroid_cosmetic_patches import SuperMetroidCosmeticPatches

AnyGameConfiguration = Union[PrimeConfiguration, EchoesConfiguration, CorruptionConfiguration,
                             SuperMetroidConfiguration]
AnyCosmeticPatches = Union[PrimeCosmeticPatches, EchoesCosmeticPatches, CorruptionCosmeticPatches,
                           SuperMetroidCosmeticPatches]

GAME_TO_CONFIGURATION: Dict[RandovaniaGame, Type[AnyGameConfiguration]] = {
    RandovaniaGame.METROID_PRIME: PrimeConfiguration,
    RandovaniaGame.METROID_PRIME_ECHOES: EchoesConfiguration,
    RandovaniaGame.METROID_PRIME_CORRUPTION: CorruptionConfiguration,
    RandovaniaGame.SUPER_METROID: SuperMetroidConfiguration,
}

GAME_TO_COSMETIC: Dict[RandovaniaGame, Type[AnyCosmeticPatches]] = {
    RandovaniaGame.METROID_PRIME: PrimeCosmeticPatches,
    RandovaniaGame.METROID_PRIME_ECHOES: EchoesCosmeticPatches,
    RandovaniaGame.METROID_PRIME_CORRUPTION: CorruptionCosmeticPatches,
    RandovaniaGame.SUPER_METROID: SuperMetroidCosmeticPatches,
}



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_translators.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetEchoesTranslators(object):
    def setupUi(self, PresetEchoesTranslators):
        if not PresetEchoesTranslators.objectName():
            PresetEchoesTranslators.setObjectName(u"PresetEchoesTranslators")
        PresetEchoesTranslators.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesTranslators)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.translators_top_layout = QGridLayout(self.centralWidget)
        self.translators_top_layout.setSpacing(6)
        self.translators_top_layout.setContentsMargins(11, 11, 11, 11)
        self.translators_top_layout.setObjectName(u"translators_top_layout")
        self.translators_top_layout.setContentsMargins(0, 4, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setFrameShadow(QFrame.Plain)
        self.scroll_area.setLineWidth(0)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.translators_scroll_contents = QWidget()
        self.translators_scroll_contents.setObjectName(u"translators_scroll_contents")
        self.translators_scroll_contents.setGeometry(QRect(0, 0, 466, 450))
        self.translators_layout = QGridLayout(self.translators_scroll_contents)
        self.translators_layout.setSpacing(6)
        self.translators_layout.setContentsMargins(11, 11, 11, 11)
        self.translators_layout.setObjectName(u"translators_layout")
        self.translators_layout.setContentsMargins(0, 0, 0, 0)
        self.translator_vanilla_colors_button = QPushButton(self.translators_scroll_contents)
        self.translator_vanilla_colors_button.setObjectName(u"translator_vanilla_colors_button")

        self.translators_layout.addWidget(self.translator_vanilla_colors_button, 2, 0, 1, 1)

        self.translator_vanilla_actual_button = QPushButton(self.translators_scroll_contents)
        self.translator_vanilla_actual_button.setObjectName(u"translator_vanilla_actual_button")

        self.translators_layout.addWidget(self.translator_vanilla_actual_button, 1, 0, 1, 1)

        self.translator_randomize_all_button = QPushButton(self.translators_scroll_contents)
        self.translator_randomize_all_button.setObjectName(u"translator_randomize_all_button")

        self.translators_layout.addWidget(self.translator_randomize_all_button, 1, 2, 1, 1)

        self.translator_randomize_all_with_unlocked_button = QPushButton(self.translators_scroll_contents)
        self.translator_randomize_all_with_unlocked_button.setObjectName(u"translator_randomize_all_with_unlocked_button")

        self.translators_layout.addWidget(self.translator_randomize_all_with_unlocked_button, 2, 2, 1, 1)

        self.translators_description = QLabel(self.translators_scroll_contents)
        self.translators_description.setObjectName(u"translators_description")
        self.translators_description.setWordWrap(True)

        self.translators_layout.addWidget(self.translators_description, 0, 0, 1, 3)

        self.scroll_area.setWidget(self.translators_scroll_contents)

        self.translators_top_layout.addWidget(self.scroll_area, 0, 0, 1, 1)

        PresetEchoesTranslators.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesTranslators)

        QMetaObject.connectSlotsByName(PresetEchoesTranslators)
    # setupUi

    def retranslateUi(self, PresetEchoesTranslators):
        PresetEchoesTranslators.setWindowTitle(QCoreApplication.translate("PresetEchoesTranslators", u"Translators Gate", None))
        self.translator_vanilla_colors_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Vanilla (Colors)", None))
        self.translator_vanilla_actual_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Vanilla (Actual)", None))
        self.translator_randomize_all_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Randomize All", None))
        self.translator_randomize_all_with_unlocked_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Randomize All with Unlocked", None))
        self.translators_description.setText(QCoreApplication.translate("PresetEchoesTranslators", u"<html><head/><body><p>Change which translator is required for all the gates in the game. Their colors are changed to match the necessary translator.</p><p>There are two vanilla options: using the actual translator requirements in the game, and using the vanilla gate colors.<br/>This is because in the original game, some translator gates are colored one way, but the translator requirement is something else.</p><p>The Emerald gate in Great Temple - Temple Sanctuary has a special case, when elevators aren't vanilla: if you enter Great Temple via Transport A, it will be permanently down.</p></body></html>", None))
    # retranslateUi




import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches


@dataclasses.dataclass(frozen=True)
class SuperMetroidCosmeticPatches(BaseCosmeticPatches):
    @classmethod
    def game(cls):
        return RandovaniaGame.SUPER_METROID



from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class RandovaniaGame(BitPackEnum, Enum):
    METROID_PRIME = "prime1"
    METROID_PRIME_ECHOES = "prime2"
    METROID_PRIME_CORRUPTION = "prime3"
    SUPER_METROID = "super_metroid"

    @property
    def short_name(self) -> str:
        return SHORT_NAME[self]

    @property
    def long_name(self) -> str:
        return LONG_NAME[self]

    @property
    def is_experimental(self) -> bool:
        return self in EXPERIMENTAL


SHORT_NAME = {
    RandovaniaGame.METROID_PRIME: "Prime",
    RandovaniaGame.METROID_PRIME_ECHOES: "Echoes",
    RandovaniaGame.METROID_PRIME_CORRUPTION: "Corruption",
    RandovaniaGame.SUPER_METROID: "SM",
}

LONG_NAME = {
    RandovaniaGame.METROID_PRIME: "Metroid Prime",
    RandovaniaGame.METROID_PRIME_ECHOES: "Metroid Prime 2: Echoes",
    RandovaniaGame.METROID_PRIME_CORRUPTION: "Metroid Prime 3: Corruption",
    RandovaniaGame.SUPER_METROID: "Super Metroid",
}

EXPERIMENTAL = {
    RandovaniaGame.METROID_PRIME_CORRUPTION,
    RandovaniaGame.SUPER_METROID,
}



import copy
import dataclasses
import functools
from typing import Dict, Optional

from PySide2 import QtWidgets, QtCore

from randovania.game_description.world.area import Area
from randovania.game_description.game_description import GameDescription
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.games.game import RandovaniaGame
from randovania.games.prime import elevators
from randovania.gui.generated.preset_elevators_ui import Ui_PresetElevators
from randovania.gui.lib import common_qt_lib, signal_handling
from randovania.gui.lib.area_list_helper import AreaListHelper
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset
from randovania.layout.lib.teleporters import TeleporterShuffleMode, TeleporterTargetList, TeleporterList
from randovania.lib import enum_lib


class PresetElevators(PresetTab, Ui_PresetElevators, AreaListHelper):
    _elevator_source_for_location: Dict[Teleporter, QtWidgets.QCheckBox]
    _elevator_source_destination: Dict[Teleporter, Optional[Teleporter]]
    _elevator_target_for_world: Dict[str, QtWidgets.QCheckBox]
    _elevator_target_for_area: Dict[int, QtWidgets.QCheckBox]

    def __init__(self, editor: PresetEditor, game: GameDescription):
        super().__init__(editor)
        self.setupUi(self)
        self.game_description = game

        self.elevator_layout.setAlignment(QtCore.Qt.AlignTop)

        for value in enum_lib.iterate_enum(TeleporterShuffleMode):
            self.elevators_combo.addItem(value.long_name, value)
        self.elevators_combo.currentIndexChanged.connect(self._update_elevator_mode)
        signal_handling.on_checked(self.skip_final_bosses_check, self._update_require_final_bosses)
        signal_handling.on_checked(self.elevators_allow_unvisited_names_check, self._update_allow_unvisited_names)

        # Elevator Source
        self._create_source_elevators()

        # Elevator Target
        self._elevator_target_for_world, self._elevator_target_for_area = self.create_area_list_selection(
            self.elevators_target_group,
            self.elevators_target_layout,
            TeleporterTargetList.areas_list(self.game_enum),
            self._on_elevator_target_check_changed,
        )

        if self.game_enum != RandovaniaGame.METROID_PRIME_ECHOES:
            self.elevators_help_sound_bug_label.setVisible(False)
            self.elevators_allow_unvisited_names_check.setVisible(False)
            self.elevators_line_3.setVisible(False)
            self.elevators_help_list_label.setVisible(False)

        if self.game_enum == RandovaniaGame.METROID_PRIME:
            self.skip_final_bosses_check.setText("Go directly to credits from Artifact Temple")
            self.skip_final_bosses_label.setText("""<html><head/><body>
            <p>Change the teleport in Artifact Temple to go directly to the credits, skipping the final bosses.</p>
            <p>This changes the requirements to <span style=" font-weight:600;">not need the final bosses</span>,
            turning certain items optional such as Plasma Beam.</p></body></html>
            """)

        elif self.game_enum == RandovaniaGame.METROID_PRIME_CORRUPTION:
            self.elevators_description_label.setText(
                self.elevators_description_label.text().replace("elevator", "teleporter")
            )

    @property
    def uses_patches_tab(self) -> bool:
        return True

    @property
    def tab_title(self):
        if self.game_enum == RandovaniaGame.METROID_PRIME_CORRUPTION:
            return "Teleporters"
        return self.windowTitle()

    @property
    def game_enum(self):
        return self.game_description.game

    def _create_check_for_source_elevator(self, location: Teleporter):
        name = elevators.get_elevator_or_area_name(self.game_enum, self.game_description.world_list, location.area_location, False)

        check = QtWidgets.QCheckBox(self.elevators_source_group)
        check.setText(name)
        check.area_location = location
        signal_handling.on_checked(check, functools.partial(self._on_elevator_source_check_changed, location))
        return check

    def _create_source_elevators(self):
        row = 0

        custom_weights = {}
        if self.game_enum == RandovaniaGame.METROID_PRIME_ECHOES:
            custom_weights = {
                2252328306: 0,  # Great Temple
                1119434212: 1,  # Agon Wastes
                1039999561: 2,  # Torvus Bog
                464164546: 3,  # Sanctuary Fortress
                1006255871: 5,  # Temple Grounds
            }
        locations = TeleporterList.areas_list(self.game_enum)
        areas: Dict[Teleporter, Area] = {
            loc: self.game_description.world_list.area_by_area_location(loc.area_location)
            for loc in locations
        }
        checks: Dict[Teleporter, QtWidgets.QCheckBox] = {
            loc: self._create_check_for_source_elevator(loc) for loc in locations
        }
        self._elevator_source_for_location = copy.copy(checks)
        self._elevator_source_destination = {}

        for location in sorted(locations,
                               key=lambda loc: (custom_weights.get(loc.world_asset_id, 0),
                                                checks[loc].text())):
            if location not in checks:
                continue

            self.elevators_source_layout.addWidget(checks.pop(location), row, 1)

            other_locations = [
                node.default_connection
                for node in areas[location].nodes
                if isinstance(node, TeleporterNode) and node.teleporter == location
            ]
            assert len(other_locations) == 1
            teleporters_in_target = [
                node.teleporter
                for node in self.game_description.world_list.area_by_area_location(other_locations[0]).nodes
                if isinstance(node, TeleporterNode)
            ]
            assert teleporters_in_target
            other_loc = teleporters_in_target[0]

            self._elevator_source_destination[location] = None

            if other_loc in checks:
                self.elevators_source_layout.addWidget(checks.pop(other_loc), row, 2)
                self._elevator_source_destination[location] = other_loc

            row += 1

    def _update_elevator_mode(self):
        with self._editor as editor:
            editor.layout_configuration_elevators = dataclasses.replace(
                editor.layout_configuration_elevators,
                mode=self.elevators_combo.currentData(),
            )

    def _update_require_final_bosses(self, checked: bool):
        with self._editor as editor:
            editor.layout_configuration_elevators = dataclasses.replace(
                editor.layout_configuration_elevators,
                skip_final_bosses=checked,
            )

    def _update_allow_unvisited_names(self, checked: bool):
        with self._editor as editor:
            editor.layout_configuration_elevators = dataclasses.replace(
                editor.layout_configuration_elevators,
                allow_unvisited_room_names=checked,
            )

    def _on_elevator_source_check_changed(self, location: Teleporter, checked: bool):
        with self._editor as editor:
            config = editor.layout_configuration_elevators
            editor.layout_configuration_elevators = dataclasses.replace(
                config,
                excluded_teleporters=config.excluded_teleporters.ensure_has_location(location, not checked),
            )

    def _on_elevator_target_check_changed(self, world_areas, checked: bool):
        with self._editor as editor:
            config = editor.layout_configuration_elevators
            editor.layout_configuration_elevators = dataclasses.replace(
                config,
                excluded_targets=config.excluded_targets.ensure_has_locations(world_areas, not checked),
            )

    def on_preset_changed(self, preset: Preset):
        config = preset.configuration

        common_qt_lib.set_combo_with_value(self.elevators_combo, config.elevators.mode)
        can_shuffle_target = config.elevators.mode not in (TeleporterShuffleMode.VANILLA,
                                                           TeleporterShuffleMode.TWO_WAY_RANDOMIZED,
                                                           TeleporterShuffleMode.TWO_WAY_UNCHECKED)
        static_areas = set(
            teleporter
            for teleporter in config.elevators.static_teleporters.keys()
        )

        for origin, destination in self._elevator_source_destination.items():
            origin_check = self._elevator_source_for_location[origin]
            dest_check = self._elevator_source_for_location.get(destination)

            is_locked = origin in static_areas
            if not can_shuffle_target:
                is_locked = is_locked or destination in static_areas

            origin_check.setEnabled(not config.elevators.is_vanilla and not is_locked)
            origin_check.setChecked(origin not in config.elevators.excluded_teleporters.locations and not is_locked)

            origin_check.setToolTip("The destination for this teleporter is locked due to other settings."
                                    if is_locked else "")

            if dest_check is None:
                if not can_shuffle_target:
                    origin_check.setEnabled(False)
                continue

            dest_check.setEnabled(can_shuffle_target and destination not in static_areas)
            if can_shuffle_target:
                dest_check.setChecked(destination not in config.elevators.excluded_teleporters.locations
                                      and destination not in static_areas)
            else:
                dest_check.setChecked(origin_check.isChecked())

        self.elevators_target_group.setEnabled(config.elevators.has_shuffled_target)
        self.skip_final_bosses_check.setChecked(config.elevators.skip_final_bosses)
        self.elevators_allow_unvisited_names_check.setChecked(config.elevators.allow_unvisited_room_names)
        self.update_area_list(
            config.elevators.excluded_targets.locations,
            True,
            self._elevator_target_for_world,
            self._elevator_target_for_area,
        )



import copy
import dataclasses
import json
import uuid
from distutils.version import StrictVersion
from enum import Enum
from pathlib import Path
from typing import Optional, TypeVar, Callable, Any, Set, List, Dict

from randovania.game_connection.memory_executor_choice import MemoryExecutorChoice
from randovania.games.game import RandovaniaGame
from randovania.interface_common import persistence, update_checker
from randovania.interface_common.persisted_options import get_persisted_options_from_data, serialized_data_for_options
from randovania.layout import game_to_class
from randovania.layout.game_to_class import AnyCosmeticPatches

T = TypeVar("T")


def identity(v: T) -> T:
    return v


class InfoAlert(Enum):
    FAQ = "faq"
    MULTI_ENERGY_ALERT = "multi-energy-alert"
    MULTIWORLD_FAQ = "multi-faq"
    NINTENDONT_UNSTABLE = "nintendont-unstable"


@dataclasses.dataclass(frozen=True)
class Serializer:
    encode: Callable[[Any], Any]
    decode: Callable[[Any], Any]


def serialize_alerts(alerts: Set[InfoAlert]) -> List[str]:
    return sorted(a.value for a in alerts)


def decode_alerts(data: List[str]):
    result = set()
    for item in data:
        try:
            result.add(InfoAlert(item))
        except ValueError:
            continue

    return result


def decode_if_not_none(value, decoder):
    if value is not None:
        return decoder(value)
    else:
        return None


@dataclasses.dataclass(frozen=True)
class PerGameOptions:
    cosmetic_patches: AnyCosmeticPatches
    input_path: Optional[Path] = None
    output_directory: Optional[Path] = None
    output_format: Optional[str] = None

    @property
    def as_json(self):
        return {
            "cosmetic_patches": self.cosmetic_patches.as_json,
            "input_path": str(self.input_path) if self.input_path is not None else None,
            "output_directory": str(self.output_directory) if self.output_directory is not None else None,
            "output_format": self.output_format if self.output_format is not None else None,
        }

    @classmethod
    def default_for_game(cls, game: RandovaniaGame) -> "PerGameOptions":
        return cls(cosmetic_patches=game_to_class.GAME_TO_COSMETIC[game]())

    @classmethod
    def from_json(cls, value, game: RandovaniaGame) -> "PerGameOptions":
        cosmetic_patches = game_to_class.GAME_TO_COSMETIC[game].from_json(value["cosmetic_patches"])
        return PerGameOptions(
            cosmetic_patches=cosmetic_patches,
            input_path=decode_if_not_none(value["input_path"], Path),
            output_format=value["output_format"],
            output_directory=decode_if_not_none(value["output_directory"], Path),
        )


def serialize_per_game_dict(val: Dict[RandovaniaGame, PerGameOptions]) -> dict:
    return {
        key.value: value.as_json
        for key, value in val.items()
    }


def decode_per_game_dict(val: dict) -> Dict[RandovaniaGame, PerGameOptions]:
    return {
        RandovaniaGame(key): PerGameOptions.from_json(value, game=RandovaniaGame(key))
        for key, value in val.items()
    }


_SERIALIZER_FOR_FIELD = {
    "last_changelog_displayed": Serializer(identity, str),
    "advanced_validate_seed_after": Serializer(identity, bool),
    "advanced_timeout_during_generation": Serializer(identity, bool),
    "auto_save_spoiler": Serializer(identity, bool),
    "dark_mode": Serializer(identity, bool),
    "experimental_games": Serializer(identity, bool),
    "selected_preset_uuid": Serializer(str, uuid.UUID),
    "per_game_options": Serializer(serialize_per_game_dict, decode_per_game_dict),
    "displayed_alerts": Serializer(serialize_alerts, decode_alerts),
    "game_backend": Serializer(lambda it: it.value, MemoryExecutorChoice),
    "nintendont_ip": Serializer(identity, str),
    "selected_tracker": Serializer(identity, str),
}


def _return_with_default(value: Optional[T], default_factory: Callable[[], T]) -> T:
    """
    Returns the given value is if it's not None, otherwise call default_factory
    :param value:
    :param default_factory:
    :return:
    """
    if value is None:
        return default_factory()
    else:
        return value


class DecodeFailedException(ValueError):
    pass


class Options:
    _data_dir: Path
    _user_dir: Path
    _on_options_changed: Optional[Callable[[], None]] = None
    _nested_autosave_level: int = 0
    _is_dirty: bool = False

    _last_changelog_displayed: str
    _advanced_validate_seed_after: Optional[bool] = None
    _advanced_timeout_during_generation: Optional[bool] = None
    _auto_save_spoiler: Optional[bool] = None
    _dark_mode: Optional[bool] = None
    _experimental_games: Optional[bool] = None
    _selected_preset_uuid: Optional[uuid.UUID] = None
    _per_game_options: Optional[Dict[RandovaniaGame, PerGameOptions]] = None
    _displayed_alerts: Optional[Set[InfoAlert]] = None
    _game_backend: Optional[MemoryExecutorChoice] = None
    _nintendont_ip: Optional[str] = None
    _selected_tracker: Optional[str] = None

    def __init__(self, data_dir: Path, user_dir: Optional[Path] = None):
        self._data_dir = data_dir
        self._user_dir = user_dir or data_dir
        self._last_changelog_displayed = str(update_checker.strict_current_version())

    @classmethod
    def with_default_data_dir(cls) -> "Options":
        return cls(persistence.local_data_dir(), persistence.roaming_data_dir())

    def _read_persisted_options(self) -> Optional[dict]:
        try:
            contents = self._data_dir.joinpath("config.json").read_text("utf-8")
            if contents.strip() == "":
                return None
            return json.loads(contents)
        except FileNotFoundError:
            return None

    def _set_field(self, field_name: str, value):
        setattr(self, "_" + field_name, value)

    def load_from_disk(self, ignore_decode_errors: bool = False) -> bool:
        """
        Loads the file created with `_save_to_disk`.
        :param ignore_decode_errors: If True, errors in the config file are ignored.
        :return: True, if a valid file exists.
        """
        try:
            persisted_data = self._read_persisted_options()

        except json.decoder.JSONDecodeError as e:
            if ignore_decode_errors:
                persisted_data = None
            else:
                raise DecodeFailedException(f"Unable to decode JSON: {e}")

        if persisted_data is None:
            return False

        persisted_options = get_persisted_options_from_data(persisted_data)
        self.load_from_persisted(persisted_options, ignore_decode_errors)
        return True

    def load_from_persisted(self,
                            persisted: dict,
                            ignore_decode_errors: bool,
                            ):
        """
        Loads fields from the given persisted options.
        :param persisted:
        :param ignore_decode_errors:
        :return:
        """
        for field_name, serializer in _SERIALIZER_FOR_FIELD.items():
            value = persisted.get(field_name, None)
            if value is not None:
                try:
                    decoded = serializer.decode(value)
                except Exception as err:
                    if ignore_decode_errors:
                        decoded = None
                    else:
                        raise DecodeFailedException(
                            f"Unable to decode field {field_name}: {err}"
                        )

                if decoded is not None:
                    self._set_field(field_name, decoded)

    def _serialize_fields(self) -> dict:
        data_to_persist = {}
        for field_name, serializer in _SERIALIZER_FOR_FIELD.items():
            value = getattr(self, "_" + field_name, None)
            if value is not None:
                data_to_persist[field_name] = serializer.encode(value)

        return serialized_data_for_options(data_to_persist)

    def _save_to_disk(self):
        """Serializes the fields of this Option and writes then to a file."""
        self._is_dirty = False
        data_to_persist = self._serialize_fields()

        self._data_dir.mkdir(parents=True, exist_ok=True)

        # Write to a separate file, so we don't corrupt the existing one in case we unexpectedly
        # are unable to finish writing the file
        new_config_path = self._data_dir.joinpath("config_new.json")
        with new_config_path.open("w") as options_file:
            json.dump(data_to_persist, options_file,
                      indent=4, separators=(',', ': '))

        # Place the new, complete, config to the desired path
        config_path = self._data_dir.joinpath("config.json")
        new_config_path.replace(config_path)

    def __enter__(self):
        self._nested_autosave_level += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._nested_autosave_level == 1:
            if self._is_dirty:
                # TODO: maybe it should be an error to change options to different values in on_options_changed?
                # previous = self._serialize_fields()
                if self._on_options_changed is not None:
                    self._on_options_changed()
                # assert previous == self._serialize_fields()
                self._save_to_disk()
        self._nested_autosave_level -= 1

    # Events
    def _set_on_options_changed(self, value):
        self._on_options_changed = value

    on_options_changed = property(fset=_set_on_options_changed)

    # Reset

    def reset_to_defaults(self):
        self._check_editable_and_mark_dirty()
        self._advanced_validate_seed_after = None
        self._advanced_timeout_during_generation = None
        self._auto_save_spoiler = None
        self._per_game_options = None
        self._displayed_alerts = None
        self._dark_mode = None
        self._experimental_games = None
        self._game_backend = None
        self._nintendont_ip = None
        self._selected_tracker = None

    # Files paths
    @property
    def internal_copies_path(self) -> Path:
        return self._data_dir.joinpath("internal_copies")

    @property
    def tracker_files_path(self) -> Path:
        return self._data_dir.joinpath("tracker")

    @property
    def presets_path(self) -> Path:
        return self._user_dir.joinpath("presets")

    @property
    def game_history_path(self) -> Path:
        return self._data_dir.joinpath("game_history")

    @property
    def logs_path(self) -> Path:
        return self._data_dir.joinpath("logs")

    @property
    def data_dir(self) -> Path:
        return self._data_dir

    @property
    def user_dir(self) -> Path:
        return self._user_dir

    # Access to Direct fields
    @property
    def last_changelog_displayed(self) -> StrictVersion:
        return StrictVersion(self._last_changelog_displayed)

    @last_changelog_displayed.setter
    def last_changelog_displayed(self, value: StrictVersion):
        if value != self.last_changelog_displayed:
            self._check_editable_and_mark_dirty()
            self._last_changelog_displayed = str(value)

    @property
    def auto_save_spoiler(self) -> bool:
        return _return_with_default(self._auto_save_spoiler, lambda: False)

    @auto_save_spoiler.setter
    def auto_save_spoiler(self, value: bool):
        self._edit_field("auto_save_spoiler", value)

    @property
    def dark_mode(self) -> bool:
        return _return_with_default(self._dark_mode, lambda: False)

    @dark_mode.setter
    def dark_mode(self, value: bool):
        self._edit_field("dark_mode", value)

    @property
    def experimental_games(self) -> bool:
        return _return_with_default(self._experimental_games, lambda: False)

    @experimental_games.setter
    def experimental_games(self, value):
        self._edit_field("experimental_games", value)

    @property
    def selected_preset_uuid(self) -> Optional[uuid.UUID]:
        return self._selected_preset_uuid

    @selected_preset_uuid.setter
    def selected_preset_uuid(self, value: uuid.UUID):
        self._edit_field("selected_preset_uuid", value)

    @property
    def game_backend(self) -> MemoryExecutorChoice:
        return _return_with_default(self._game_backend, lambda: MemoryExecutorChoice.DOLPHIN)

    @game_backend.setter
    def game_backend(self, value: MemoryExecutorChoice):
        self._edit_field("game_backend", value)

    @property
    def nintendont_ip(self) -> Optional[str]:
        return self._nintendont_ip

    @nintendont_ip.setter
    def nintendont_ip(self, value: Optional[str]):
        self._edit_field("nintendont_ip", value)

    @property
    def selected_tracker(self) -> str:
        return self._selected_tracker

    @selected_tracker.setter
    def selected_tracker(self, value: str):
        self._edit_field("selected_tracker", value)

    @property
    def displayed_alerts(self):
        return _return_with_default(self._displayed_alerts, set)

    @displayed_alerts.setter
    def displayed_alerts(self, value):
        self._edit_field("displayed_alerts", value)

    def is_alert_displayed(self, value: InfoAlert):
        return value in self.displayed_alerts

    def mark_alert_as_displayed(self, value: InfoAlert):
        if value not in self.displayed_alerts:
            # Create a copy so we don't modify the existing field
            alerts = set(self.displayed_alerts)
            alerts.add(value)
            with self:
                self.displayed_alerts = alerts

    @property
    def per_game_options(self):
        return _return_with_default(self._per_game_options, lambda: {})

    @per_game_options.setter
    def per_game_options(self, value):
        self._edit_field("per_game_options", value)

    def options_for_game(self, game: RandovaniaGame) -> PerGameOptions:
        return self.per_game_options.get(game, PerGameOptions.default_for_game(game))

    def set_options_for_game(self, game: RandovaniaGame, per_game: PerGameOptions):
        per_game_options = copy.copy(self.per_game_options)
        per_game_options[game] = per_game
        self.per_game_options = per_game_options

    # Advanced

    @property
    def advanced_validate_seed_after(self) -> bool:
        return _return_with_default(self._advanced_validate_seed_after, lambda: True)

    @advanced_validate_seed_after.setter
    def advanced_validate_seed_after(self, value: bool):
        self._edit_field("advanced_validate_seed_after", value)

    @property
    def advanced_timeout_during_generation(self) -> bool:
        return _return_with_default(self._advanced_timeout_during_generation, lambda: True)

    @advanced_timeout_during_generation.setter
    def advanced_timeout_during_generation(self, value: bool):
        self._edit_field("advanced_timeout_during_generation", value)

    ######

    def _check_editable_and_mark_dirty(self):
        """Checks if _nested_autosave_level is not 0 and marks at least one value was changed."""
        assert self._nested_autosave_level != 0, "Attempting to edit an Options, but it wasn't made editable"
        self._is_dirty = True

    def _edit_field(self, field_name: str, new_value):
        current_value = getattr(self, field_name)
        if current_value != new_value:
            self._check_editable_and_mark_dirty()
            self._set_field(field_name, new_value)



import dataclasses
from random import Random
from unittest.mock import MagicMock, patch, call, ANY

import pytest

from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    HintDarkTemple
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.search import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world.teleporter import Teleporter
from randovania.games.game import RandovaniaGame
from randovania.generator import base_patches_factory
from randovania.layout.lib.teleporters import TeleporterShuffleMode
from randovania.layout.prime2.translator_configuration import LayoutTranslatorRequirement


@pytest.mark.parametrize("skip_final_bosses", [False, True])
def test_add_elevator_connections_to_patches_vanilla(echoes_game_description,
                                                     skip_final_bosses: bool,
                                                     default_layout_configuration):
    # Setup
    expected = dataclasses.replace(echoes_game_description.create_game_patches())
    if skip_final_bosses:
        expected.elevator_connection[Teleporter(0x3bfa3eff, 0x87d35ee4, 0x82a008b)] = AreaLocation(1006255871,
                                                                                                   1393588666)
    config = default_layout_configuration
    config = dataclasses.replace(config,
                                 elevators=dataclasses.replace(config.elevators,
                                                               skip_final_bosses=skip_final_bosses))

    # Run
    result = base_patches_factory.add_elevator_connections_to_patches(
        config,
        Random(0),
        echoes_game_description.create_game_patches())

    # Assert
    assert result == expected


@pytest.mark.parametrize("skip_final_bosses", [False, True])
def test_add_elevator_connections_to_patches_random(echoes_game_description,
                                                    skip_final_bosses: bool,
                                                    default_layout_configuration):
    # Setup
    game = echoes_game_description
    layout_configuration = dataclasses.replace(
        default_layout_configuration,
        elevators=dataclasses.replace(
            default_layout_configuration.elevators,
            mode=TeleporterShuffleMode.TWO_WAY_RANDOMIZED,
            skip_final_bosses=skip_final_bosses,
        ),
    )
    expected = dataclasses.replace(
        game.create_game_patches(),
        elevator_connection={
            Teleporter(0x3bfa3eff, 0xaded752e, 0x9001b): AreaLocation(1039999561, 1868895730),
            Teleporter(0x3bfa3eff, 0x62ff94ee, 0x180086): AreaLocation(1039999561, 3479543630),
            Teleporter(0x3bfa3eff, 0xac32f338, 0x1e000d): AreaLocation(2252328306, 408633584),
            Teleporter(0x3bfa3eff, 0x4cc37f4a, 0x200063): AreaLocation(1119434212, 3331021649),
            Teleporter(0x3bfa3eff, 0x87d35ee4, 0x82a008b): (AreaLocation(1006255871, 1393588666)
                                                            if skip_final_bosses else AreaLocation(2252328306,
                                                                                                   2068511343)),
            Teleporter(0x863fcd72, 0x7b4afa6f, 0x9007d): AreaLocation(0x3BFA3EFF, 0x87D35EE4),
            Teleporter(0x3bfa3eff, 0xcdf7686b, 0x33006e): AreaLocation(1039999561, 3205424168),
            Teleporter(0x3bfa3eff, 0x503a0640, 0x36001f): AreaLocation(1119434212, 2806956034),
            Teleporter(0x863fcd72, 0x185b40f0, 0x98): AreaLocation(1006255871, 2889020216),
            Teleporter(0x863fcd72, 0x9860cbb0, 0x6002c): AreaLocation(464164546, 3145160350),
            Teleporter(0x863fcd72, 0x8f01b104, 0x80021): AreaLocation(464164546, 900285955),
            Teleporter(0x42b935e4, 0x57ce3a52, 0x7a): AreaLocation(464164546, 3528156989),
            Teleporter(0x42b935e4, 0xa74ec002, 0x13007b): AreaLocation(1006255871, 1345979968),
            Teleporter(0x42b935e4, 0xc68b5b51, 0x2d0073): AreaLocation(1006255871, 1287880522),
            Teleporter(0x3dfd2249, 0x6f6515f2, 0x81): AreaLocation(1006255871, 2918020398),
            Teleporter(0x3dfd2249, 0xcf659f4e, 0x21008a): AreaLocation(1006255871, 1660916974),
            Teleporter(0x3dfd2249, 0xbf0ee428, 0x450030): AreaLocation(1006255871, 3455543403),
            Teleporter(0x1baa96c2, 0xd24b673d, 0x26): AreaLocation(1119434212, 1473133138),
            Teleporter(0x1baa96c2, 0x35a94603, 0x130094): AreaLocation(2252328306, 2399252740),
            Teleporter(0x1baa96c2, 0xbb77569e, 0x190087): AreaLocation(2252328306, 2556480432),
            Teleporter(0x1baa96c2, 0xbaf94a13, 0xc36007c): AreaLocation(464164546, 1564082177),
            Teleporter(0x1baa96c2, 0x5d3a0001, 0x41010a): AreaLocation(464164546, 3136899603),
        })

    # Run
    result = base_patches_factory.add_elevator_connections_to_patches(
        layout_configuration,
        Random(0),
        game.create_game_patches(),
    )

    # Assert
    assert len(result.elevator_connection) == len(expected.elevator_connection)
    assert result.elevator_connection == expected.elevator_connection

    assert result == expected


def test_gate_assignment_for_configuration_all_emerald(echoes_resource_database):
    # Setup
    emerald = find_resource_info_with_long_name(echoes_resource_database.item, "Emerald Translator")
    indices = [1, 15, 23]

    configuration = MagicMock()
    configuration.translator_configuration.translator_requirement = {
        TranslatorGate(index): LayoutTranslatorRequirement.EMERALD
        for index in indices
    }

    rng = MagicMock()

    # Run
    results = base_patches_factory.gate_assignment_for_configuration(configuration, echoes_resource_database, rng)

    # Assert
    assert results == {
        TranslatorGate(index): emerald
        for index in indices
    }


def test_gate_assignment_for_configuration_all_random(echoes_resource_database):
    # Setup
    violet = find_resource_info_with_long_name(echoes_resource_database.item, "Violet Translator")
    emerald = find_resource_info_with_long_name(echoes_resource_database.item, "Emerald Translator")

    configuration = MagicMock()
    configuration.translator_configuration.translator_requirement = {
        TranslatorGate(index): LayoutTranslatorRequirement.RANDOM
        for index in [1, 15, 23]
    }

    rng = MagicMock()
    rng.choice.side_effect = [
        LayoutTranslatorRequirement.EMERALD,
        LayoutTranslatorRequirement.VIOLET,
        LayoutTranslatorRequirement.EMERALD,
    ]

    # Run
    results = base_patches_factory.gate_assignment_for_configuration(configuration, echoes_resource_database, rng)

    # Assert
    assert results == {
        TranslatorGate(1): emerald,
        TranslatorGate(15): violet,
        TranslatorGate(23): emerald,
    }


@pytest.mark.parametrize("is_multiworld", [False, True])
def test_add_default_hints_to_patches(echoes_game_description, empty_patches, is_multiworld):
    # Setup
    rng = MagicMock()

    def _light_suit_location_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.LIGHT_SUIT_LOCATION,
                                                     HintItemPrecision.DETAILED, include_owner=False),
                    PickupIndex(number))

    def _guardian_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.GUARDIAN,
                                                     HintItemPrecision.DETAILED, include_owner=False),
                    PickupIndex(number))

    def _keybearer_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.KEYBEARER,
                                                     HintItemPrecision.BROAD_CATEGORY,
                                                     include_owner=True),
                    PickupIndex(number))

    expected = {
        # Keybearer
        LogbookAsset(0xE3B417BF): _keybearer_hint(11),
        LogbookAsset(0x65206511): _keybearer_hint(15),
        LogbookAsset(0x28E8C41A): _keybearer_hint(19),
        # Agon
        LogbookAsset(0x150E8DB8): _keybearer_hint(45),
        LogbookAsset(0xDE525E1D): _keybearer_hint(53),
        # Torvus
        LogbookAsset(0x58C62CB3): _keybearer_hint(68),
        LogbookAsset(0x939AFF16): _keybearer_hint(91),
        # Sanctuary
        LogbookAsset(0x62CC4DC3): _keybearer_hint(117),
        LogbookAsset(0xA9909E66): _keybearer_hint(106),

        # Locations with hints
        LogbookAsset(1041207119): _light_suit_location_hint(24),
        LogbookAsset(4115881194): _guardian_hint(43),
        LogbookAsset(1948976790): _guardian_hint(79),
        LogbookAsset(3212301619): _guardian_hint(115),

        # Dark Temple hints
        LogbookAsset(67497535): Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.AGON_WASTES),
        LogbookAsset(4072633400): Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.TORVUS_BOG),
        LogbookAsset(0x82919C91): Hint(HintType.RED_TEMPLE_KEY_SET, None,
                                       dark_temple=HintDarkTemple.SANCTUARY_FORTRESS),

        # Jokes
        LogbookAsset(0x49CD4F34): Hint(HintType.JOKE, None),
        LogbookAsset(0x9F94AC29): Hint(HintType.JOKE, None),
    }

    # Run
    result = base_patches_factory.add_echoes_default_hints_to_patches(
        rng, empty_patches, echoes_game_description.world_list, num_joke=2, is_multiworld=is_multiworld)

    # Assert
    rng.shuffle.assert_has_calls([call(ANY), call(ANY)])
    assert result.hints == expected


@patch("randovania.generator.base_patches_factory.add_echoes_default_hints_to_patches", autospec=True)
@patch("randovania.generator.base_patches_factory.starting_location_for_configuration", autospec=True)
@patch("randovania.generator.base_patches_factory.gate_assignment_for_configuration", autospec=True)
@patch("randovania.generator.base_patches_factory.add_elevator_connections_to_patches", autospec=True)
def test_create_base_patches(mock_add_elevator_connections_to_patches: MagicMock,
                             mock_gate_assignment_for_configuration: MagicMock,
                             mock_starting_location_for_config: MagicMock,
                             mock_add_default_hints_to_patches: MagicMock,
                             mocker,
                             ):
    # Setup
    rng = MagicMock()
    game = MagicMock()
    layout_configuration = MagicMock()
    layout_configuration.game = RandovaniaGame.METROID_PRIME_ECHOES
    mock_replace: MagicMock = mocker.patch("dataclasses.replace")
    is_multiworld = MagicMock()

    patches = ([
        game.create_game_patches.return_value,
        mock_replace.return_value,
        mock_add_elevator_connections_to_patches.return_value,
    ])
    patches.append(patches[-1].assign_gate_assignment.return_value)
    patches.append(patches[-1].assign_starting_location.return_value)

    # Run
    result = base_patches_factory.create_base_patches(layout_configuration, rng, game, is_multiworld, player_index=0)

    # Assert
    game.create_game_patches.assert_called_once_with()
    mock_replace.assert_called_once_with(game.create_game_patches.return_value, player_index=0)
    mock_add_elevator_connections_to_patches.assert_called_once_with(layout_configuration, rng, patches[1])

    # Gate Assignment
    mock_gate_assignment_for_configuration.assert_called_once_with(layout_configuration, game.resource_database, rng)
    patches[2].assign_gate_assignment.assert_called_once_with(mock_gate_assignment_for_configuration.return_value)

    # Starting Location
    mock_starting_location_for_config.assert_called_once_with(layout_configuration, game, rng)
    patches[3].assign_starting_location.assert_called_once_with(mock_starting_location_for_config.return_value)

    # Hints
    mock_add_default_hints_to_patches.assert_called_once_with(rng, patches[4], game.world_list, num_joke=2,
                                                              is_multiworld=is_multiworld)

    assert result is mock_add_default_hints_to_patches.return_value



import pytest
from mock import MagicMock

from randovania.game_connection.connection_base import InventoryItem
from randovania.game_connection.connector.prime1_remote_connector import Prime1RemoteConnector
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.games.game import RandovaniaGame
from randovania.games.prime.prime1_dol_patches import Prime1DolVersion


@pytest.fixture(name="version")
def prime1_version():
    from randovania.games.prime import prime1_dol_versions
    return prime1_dol_versions.ALL_VERSIONS[0]


@pytest.fixture(name="connector")
def remote_connector(version: Prime1DolVersion):
    connector = Prime1RemoteConnector(version)
    return connector


@pytest.mark.parametrize("artifact", [False, True])
@pytest.mark.asyncio
async def test_patches_for_pickup(connector: Prime1RemoteConnector, mocker, artifact: bool, generic_item_category):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.adjust_item_amount_and_capacity_patch")
    mock_increment_capacity: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.increment_item_capacity_patch")
    mock_artifact_layer: MagicMock = mocker.patch(
        "randovania.games.prime.prime1_dol_patches.set_artifact_layer_active_patch")

    db = connector.game.resource_database
    if artifact:
        extra = (db.get_item(30), 1)
    else:
        extra = (db.energy_tank, db.energy_tank.max_capacity)

    pickup = PickupEntry("Pickup", 0, generic_item_category, generic_item_category, progression=tuple(),
                         extra_resources=(
                             extra,
                         ))
    inventory = {
        db.multiworld_magic_item: InventoryItem(0, 0),
        db.energy_tank: InventoryItem(1, 1),
    }

    # Run
    patches, message = await connector._patches_for_pickup("Someone", pickup, inventory)

    # Assert
    expected_patches = []

    if artifact:
        mock_artifact_layer.assert_called_once_with(connector.version, 2, True)
        expected_patches.append(mock_artifact_layer.return_value)
        used_patch, unused_patch = mock_increment_capacity, mock_item_patch
    else:
        mock_artifact_layer.assert_not_called()
        used_patch, unused_patch = mock_item_patch, mock_increment_capacity

    expected_patches.insert(0, used_patch.return_value)
    used_patch.assert_called_once_with(connector.version.powerup_functions,
                                       RandovaniaGame.METROID_PRIME,
                                       extra[0].index,
                                       extra[1])
    unused_patch.assert_not_called()

    assert patches == expected_patches
    assert message == "Received Pickup from Someone."



from unittest.mock import MagicMock

from randovania.interface_common import simplified_patcher
from randovania.interface_common.options import Options
from randovania.layout.permalink import Permalink


def test_generate_layout(mocker):
    # Setup
    options: Options = MagicMock()
    permalink: Permalink = MagicMock()
    progress_update = MagicMock()

    mock_generate_layout = mocker.patch("randovania.interface_common.echoes.generate_description", autospec=True)
    mock_constant_percentage_callback = mocker.patch(
        "randovania.interface_common.simplified_patcher.ConstantPercentageCallback",
        autospec=False,  # TODO: pytest-qt bug
    )

    # Run
    simplified_patcher.generate_layout(options, permalink, progress_update)

    # Assert
    mock_constant_percentage_callback.assert_called_once_with(progress_update, -1)
    mock_generate_layout.assert_called_once_with(
        permalink=permalink,
        status_update=mock_constant_percentage_callback.return_value,
        validate_after_generation=options.advanced_validate_seed_after,
        timeout_during_generation=options.advanced_timeout_during_generation,
        attempts=None,
    )



import asyncio
import platform
import sys
from unittest.mock import patch, MagicMock, call

import pytest

from randovania.games.patchers import csharp_subprocess


@pytest.fixture(
    params=[False, True],
    name="mock_is_windows")
def _mock_is_windows(request):
    with patch("randovania.games.patchers.csharp_subprocess.is_windows", return_value=request.param):
        yield request.param


@pytest.mark.parametrize(["system_name", "expected"], [
    ("Windows", True),
    ("Linux", False),
])
def test_is_windows(mocker, system_name, expected):
    # Setup
    mocker.patch("platform.system", return_value=system_name)

    # Run
    result = csharp_subprocess.is_windows()

    # Assert
    assert result == expected


def test_process_command_no_thread(echo_tool, mock_is_windows, mocker, monkeypatch):
    read_callback = MagicMock()

    csharp_subprocess.IO_LOOP = None
    if platform.system() == "Windows":
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    mock_set_event: MagicMock = mocker.patch("asyncio.set_event_loop_policy")
    loop_policy = MagicMock()
    monkeypatch.setattr(asyncio, "WindowsProactorEventLoopPolicy", loop_policy, raising=False)

    # Run
    csharp_subprocess.process_command(
        [
            sys.executable,
            str(echo_tool)
        ],
        "hello\r\nthis is a nice world\r\n\r\nWe some crazy stuff.",
        read_callback,
        add_mono_if_needed=False,
    )

    # Assert
    read_callback.assert_has_calls([
        call("hello"),
        call("this is a nice world"),
        call("We some crazy stuff."),
    ])
    if mock_is_windows:
        mock_set_event.assert_called_once_with(loop_policy.return_value)
    else:
        mock_set_event.assert_not_called()



from unittest.mock import MagicMock

import pytest
from mock import AsyncMock

from randovania.game_connection.executor.dolphin_executor import DolphinExecutor
from randovania.game_connection.game_connection import GameConnection


@pytest.fixture(name="connection")
def _connection(skip_qtbot):
    return GameConnection(MagicMock())


@pytest.mark.asyncio
async def test_auto_update(skip_qtbot, qapp):
    # Setup
    update_mock = AsyncMock()
    backend = MagicMock()

    game_connection = GameConnection(backend)
    game_connection._notify_status = MagicMock()
    game_connection.update = update_mock

    # Run
    await game_connection._auto_update()

    # Assert
    update_mock.assert_awaited_once_with(game_connection._dt)
    game_connection._notify_status.assert_called_once_with()


def test_pretty_current_status(skip_qtbot):
    # Setup
    connection = GameConnection(DolphinExecutor())

    # Run
    result = connection.pretty_current_status

    # Assert
    assert result == f"Dolphin: Disconnected"



import copy
import dataclasses
from random import Random

from randovania.game_description import default_database
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.assignment import GateAssignment
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    HintDarkTemple
from randovania.game_description.world.node import LogbookNode, LoreType
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.generator import elevator_distributor
from randovania.lib.enum_lib import iterate_enum
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration
from randovania.layout.lib.teleporters import TeleporterShuffleMode
from randovania.layout.prime2.translator_configuration import LayoutTranslatorRequirement


class MissingRng(Exception):
    pass


def add_elevator_connections_to_patches(configuration: EchoesConfiguration,
                                        rng: Random,
                                        patches: GamePatches) -> GamePatches:
    """
    :param configuration:
    :param rng:
    :param patches:
    :return:
    """
    elevator_connection = copy.copy(patches.elevator_connection)
    elevators = configuration.elevators

    if not elevators.is_vanilla:
        if rng is None:
            raise MissingRng("Elevator")

        world_list = default_database.game_description_for(configuration.game).world_list
        elevator_db = elevator_distributor.create_elevator_database(
            world_list, elevators.editable_teleporters)

        if elevators.mode in {TeleporterShuffleMode.TWO_WAY_RANDOMIZED, TeleporterShuffleMode.TWO_WAY_UNCHECKED}:
            connections = elevator_distributor.two_way_elevator_connections(
                rng=rng,
                elevator_database=elevator_db,
                between_areas=elevators.mode == TeleporterShuffleMode.TWO_WAY_RANDOMIZED
            )
        else:
            connections = elevator_distributor.one_way_elevator_connections(
                rng=rng,
                elevator_database=elevator_db,
                target_locations=elevators.valid_targets,
                replacement=elevators.mode == TeleporterShuffleMode.ONE_WAY_ELEVATOR_REPLACEMENT,
            )

        elevator_connection.update(connections)

    for teleporter, destination in elevators.static_teleporters.items():
        elevator_connection[teleporter] = destination

    return dataclasses.replace(patches, elevator_connection=elevator_connection)


def gate_assignment_for_configuration(configuration: EchoesConfiguration,
                                      resource_database: ResourceDatabase,
                                      rng: Random,
                                      ) -> GateAssignment:
    """
    :param configuration:
    :param resource_database:
    :param rng:
    :return:
    """

    all_choices = list(LayoutTranslatorRequirement)
    all_choices.remove(LayoutTranslatorRequirement.RANDOM)
    all_choices.remove(LayoutTranslatorRequirement.RANDOM_WITH_REMOVED)
    without_removed = copy.copy(all_choices)
    without_removed.remove(LayoutTranslatorRequirement.REMOVED)
    random_requirements = {LayoutTranslatorRequirement.RANDOM, LayoutTranslatorRequirement.RANDOM_WITH_REMOVED}

    result = {}
    for gate, requirement in configuration.translator_configuration.translator_requirement.items():
        if requirement in random_requirements:
            if rng is None:
                raise MissingRng("Translator")
            requirement = rng.choice(all_choices if requirement == LayoutTranslatorRequirement.RANDOM_WITH_REMOVED
                                     else without_removed)

        result[gate] = resource_database.get_by_type_and_index(ResourceType.ITEM, requirement.item_index)

    return result


def starting_location_for_configuration(configuration: EchoesConfiguration,
                                        game: GameDescription,
                                        rng: Random,
                                        ) -> AreaLocation:
    locations = list(configuration.starting_location.locations)
    if len(locations) == 0:
        raise ValueError("No available starting locations")
    elif len(locations) == 1:
        location = locations[0]
    else:
        if rng is None:
            raise MissingRng("Starting Location")
        location = rng.choice(locations)

    return location


def add_echoes_default_hints_to_patches(rng: Random,
                                        patches: GamePatches,
                                        world_list: WorldList,
                                        num_joke: int,
                                        is_multiworld: bool,
                                        ) -> GamePatches:
    """
    Adds hints that are present on all games.
    :param rng:
    :param patches:
    :param world_list:
    :param num_joke
    :param is_multiworld
    :return:
    """

    for node in world_list.all_nodes:
        if isinstance(node, LogbookNode) and node.lore_type == LoreType.LUMINOTH_WARRIOR:
            patches = patches.assign_hint(node.resource(),
                                          Hint(HintType.LOCATION,
                                               PrecisionPair(HintLocationPrecision.KEYBEARER,
                                                             HintItemPrecision.BROAD_CATEGORY,
                                                             include_owner=True),
                                               PickupIndex(node.hint_index)))

    all_logbook_assets = [node.resource()
                          for node in world_list.all_nodes
                          if isinstance(node, LogbookNode)
                          and node.resource() not in patches.hints
                          and node.lore_type.holds_generic_hint]

    rng.shuffle(all_logbook_assets)

    # The 4 guaranteed hints
    indices_with_hint = [
        (PickupIndex(24), HintLocationPrecision.LIGHT_SUIT_LOCATION),  # Light Suit
        (PickupIndex(43), HintLocationPrecision.GUARDIAN),  # Dark Suit (Amorbis)
        (PickupIndex(79), HintLocationPrecision.GUARDIAN),  # Dark Visor (Chykka)
        (PickupIndex(115), HintLocationPrecision.GUARDIAN),  # Annihilator Beam (Quadraxis)
    ]
    rng.shuffle(indices_with_hint)
    for index, location_type in indices_with_hint:
        if not all_logbook_assets:
            break

        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.LOCATION,
                                                          PrecisionPair(location_type, HintItemPrecision.DETAILED,
                                                                        include_owner=False),
                                                          index))

    # Dark Temple hints
    temple_hints = list(iterate_enum(HintDarkTemple))
    while all_logbook_assets and temple_hints:
        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.RED_TEMPLE_KEY_SET, None,
                                                          dark_temple=temple_hints.pop(0)))

    # Jokes
    while num_joke > 0 and all_logbook_assets:
        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.JOKE, None))
        num_joke -= 1

    return patches


def create_base_patches(configuration: EchoesConfiguration,
                        rng: Random,
                        game: GameDescription,
                        is_multiworld: bool,
                        player_index: int,
                        ) -> GamePatches:
    """
    """
    patches = dataclasses.replace(game.create_game_patches(),
                                  player_index=player_index)


    if hasattr(configuration, "elevators"):
        patches = add_elevator_connections_to_patches(configuration, rng, patches)

    # Gates
    if configuration.game == RandovaniaGame.METROID_PRIME_ECHOES:
        patches = patches.assign_gate_assignment(
            gate_assignment_for_configuration(configuration, game.resource_database, rng))

    # Starting Location
    patches = patches.assign_starting_location(
        starting_location_for_configuration(configuration, game, rng))

    # Hints
    if rng is not None and configuration.game == RandovaniaGame.METROID_PRIME_ECHOES:
        patches = add_echoes_default_hints_to_patches(rng, patches, game.world_list,
                                                      num_joke=2, is_multiworld=is_multiworld)

    return patches



import copy
import dataclasses
import json
from unittest.mock import MagicMock, patch, ANY

import pytest

import randovania
from randovania.game_description import default_database
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.default_database import default_prime2_memo_data
from randovania.game_description.resources.pickup_entry import PickupModel, ConditionalResources
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world.teleporter import Teleporter
from randovania.games.game import RandovaniaGame
from randovania.games.patchers import claris_patcher_file
from randovania.games.prime.patcher_file_lib import pickup_exporter
from randovania.generator.item_pool import pickup_creator, pool_creator
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.layout.prime2.hint_configuration import SkyTempleKeyHintMode, HintConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.base.major_item_state import MajorItemState
from randovania.layout.base.pickup_model import PickupModelStyle
from randovania.layout.lib.teleporters import TeleporterShuffleMode


def test_add_header_data_to_result():
    # Setup
    description = MagicMock()
    description.shareable_word_hash = "<shareable_word_hash>"
    description.shareable_hash = "<shareable_hash>"
    expected = {
        "permalink": "-permalink-",
        "seed_hash": "- <shareable_word_hash> (<shareable_hash>)",
        "randovania_version": randovania.VERSION,
        'shareable_hash': '<shareable_hash>',
        'shareable_word_hash': '<shareable_word_hash>',
    }
    result = {}

    # Run
    claris_patcher_file._add_header_data_to_result(description, result)

    # Assert
    assert json.loads(json.dumps(result)) == expected


def test_create_spawn_point_field(echoes_resource_database, empty_patches):
    # Setup
    patches = empty_patches.assign_starting_location(AreaLocation(100, 5000)).assign_extra_initial_items({
        echoes_resource_database.get_by_type_and_index(ResourceType.ITEM, 15): 3,
    })

    capacities = [
        {'amount': 3 if item.index == 15 else 0, 'index': item.index}
        for item in echoes_resource_database.item
    ]

    # Run
    result = claris_patcher_file._create_spawn_point_field(patches, echoes_resource_database)

    # Assert
    assert result == {
        "location": {
            "world_asset_id": 100,
            "area_asset_id": 5000,
        },
        "amount": capacities,
        "capacity": capacities,
    }


def test_create_elevators_field_no_elevator(empty_patches, echoes_game_description):
    # Setup
    # Run
    with pytest.raises(ValueError) as exp:
        claris_patcher_file._create_elevators_field(empty_patches, echoes_game_description)

    # Assert
    assert str(exp.value) == "Invalid elevator count. Expected 22, got 0."


@pytest.mark.parametrize("vanilla_gateway", [False, True])
def test_create_elevators_field_elevators_for_a_seed(vanilla_gateway: bool,
                                                     echoes_game_description, empty_patches):
    # Setup
    patches = echoes_game_description.create_game_patches()

    elevator_connection = copy.copy(patches.elevator_connection)
    elevator_connection[Teleporter(0x3BFA3EFF, 0xADED752E, 0x9001B)] = AreaLocation(464164546, 900285955)
    elevator_connection[Teleporter(0x3BFA3EFF, 0x62FF94EE, 0x180086)] = AreaLocation(1039999561, 3479543630)

    if not vanilla_gateway:
        elevator_connection[Teleporter(0x3BFA3EFF, 0x87D35EE4, 0x82A008B)] = AreaLocation(2252328306, 3619928121)

    patches = dataclasses.replace(patches, elevator_connection=elevator_connection)

    # Run
    result = claris_patcher_file._create_elevators_field(patches, echoes_game_description)

    # Assert
    expected = [
        {"instance_id": 589851,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2918020398},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "room_name": "Transport to Sanctuary Spider side"},

        {"instance_id": 1572998,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1660916974},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "room_name": "Transport to Torvus Temple Access"},

        {"instance_id": 1966093,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2889020216},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 1868895730},
         "room_name": "Transport to Torvus Entrance"},

        {"instance_id": 2097251,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1287880522},
         "target_location": {"world_asset_id": 2252328306, "area_asset_id": 2399252740},
         "room_name": "Transport to Temple Transport Violet"},

        {"instance_id": 136970379,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2278776548},
         "target_location": {"world_asset_id": 2252328306,
                             "area_asset_id": 2068511343 if vanilla_gateway else 3619928121},
         "room_name": "Sky Temple Gateway" if vanilla_gateway else "Transport to Sanctum"},

        {"instance_id": 3342446,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 3455543403},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 3528156989},
         "room_name": "Transport to Sanctuary Entrance"},

        {"instance_id": 3538975,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1345979968},
         "target_location": {"world_asset_id": 2252328306, "area_asset_id": 408633584},
         "room_name": "Transport to Temple Transport Emerald"},

        {"instance_id": 152,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 408633584},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1345979968},
         "room_name": "Transport to Sanctuary Quadrant"},

        {"instance_id": 393260,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 2556480432},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 2918020398},
         "room_name": "Transport to Torvus Quadrant"},

        {"instance_id": 524321,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 2399252740},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1287880522},
         "room_name": "Transport to Agon Quadrant"},

        {'instance_id': 589949,
         'origin_location': {'area_asset_id': 2068511343, 'world_asset_id': 2252328306},
         'target_location': {'area_asset_id': 2278776548, 'world_asset_id': 1006255871},
         'room_name': 'Sky Temple Energy Controller'},

        {"instance_id": 122,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 1473133138},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1660916974},
         "room_name": "Transport to Agon Gate"},

        {"instance_id": 1245307,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 2806956034},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "room_name": "Transport to Torvus Temple Access"},

        {"instance_id": 2949235,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 3331021649},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "room_name": "Transport to Sanctuary Spider side"},

        {"instance_id": 129,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 1868895730},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 2889020216},
         "room_name": "Transport to Torvus Gate"},

        {"instance_id": 2162826,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "target_location": {"world_asset_id": 1119434212, "area_asset_id": 2806956034},
         "room_name": "Transport to Agon Portal Access"},

        {"instance_id": 4522032,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 3205424168},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 3145160350},
         "room_name": "Transport to Sanctuary Vault side"},

        {"instance_id": 38,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 3528156989},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 3455543403},
         "room_name": "Transport to Sanctuary Gate"},

        {"instance_id": 1245332,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "target_location": {"world_asset_id": 1119434212, "area_asset_id": 3331021649},
         "room_name": "Transport to Agon Temple Access"},

        {"instance_id": 1638535,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 3145160350},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3205424168},
         "room_name": "Transport to Lower Torvus Access"},

        {'instance_id': 204865660,
         'origin_location': {'world_asset_id': 464164546, 'area_asset_id': 3136899603},
         'target_location': {'world_asset_id': 464164546, 'area_asset_id': 1564082177},
         'room_name': 'Aerie Transport Station'},

        {'instance_id': 4260106,
         'origin_location': {'world_asset_id': 464164546, 'area_asset_id': 1564082177},
         'target_location': {'world_asset_id': 464164546, 'area_asset_id': 3136899603},
         'room_name': 'Aerie'}
    ]
    assert result == expected


def test_create_translator_gates_field():
    # Setup
    gate_assignment = {
        TranslatorGate(1): SimpleResourceInfo(10, "LongA", "A", ResourceType.ITEM),
        TranslatorGate(3): SimpleResourceInfo(50, "LongB", "B", ResourceType.ITEM),
        TranslatorGate(4): SimpleResourceInfo(10, "LongA", "A", ResourceType.ITEM),
    }

    # Run
    result = claris_patcher_file._create_translator_gates_field(gate_assignment)

    # Assert
    assert result == [
        {"gate_index": 1, "translator_index": 10},
        {"gate_index": 3, "translator_index": 50},
        {"gate_index": 4, "translator_index": 10},
    ]


@pytest.mark.parametrize("elevators", [TeleporterShuffleMode.VANILLA, TeleporterShuffleMode.TWO_WAY_RANDOMIZED])
def test_apply_translator_gate_patches(elevators):
    # Setup
    target = {}

    # Run
    claris_patcher_file._apply_translator_gate_patches(target, elevators)

    # Assert
    assert target == {
        "always_up_gfmc_compound": True,
        "always_up_torvus_temple": True,
        "always_up_great_temple": elevators != TeleporterShuffleMode.VANILLA,
    }


def test_get_single_hud_text_locked_pbs():
    # Run
    result = pickup_exporter._get_single_hud_text("Locked Power Bomb Expansion",
                                                  claris_patcher_file._simplified_memo_data(),
                                                  tuple())

    # Assert
    assert result == "Power Bomb Expansion acquired, but the main Power Bomb is required to use it."


def test_pickup_data_for_seeker_launcher(echoes_item_database, echoes_resource_database):
    # Setup
    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=0,
        included_ammo=(5,),
    )
    pickup = pickup_creator.create_major_item(
        echoes_item_database.major_items["Seeker Launcher"],
        state,
        True,
        echoes_resource_database,
        echoes_item_database.ammo["Missile Expansion"],
        True
    )
    creator = pickup_exporter.PickupExporterSolo(claris_patcher_file._simplified_memo_data())

    # Run
    details = creator.export(PickupIndex(0), PickupTarget(pickup, 0), pickup, PickupModelStyle.ALL_VISIBLE)
    result = claris_patcher_file.echoes_pickup_details_to_patcher(
        details, MagicMock())

    # Assert
    assert result == {
        "pickup_index": 0,
        "scan": "Seeker Launcher",
        "model_name": "SeekerLauncher",
        "hud_text": ["Seeker Launcher acquired, but the Missile Launcher is required to use it.",
                     "Seeker Launcher acquired!"],
        'resources': [{'amount': 5, 'index': 71},
                      {'amount': 1, 'index': 47},
                      {'amount': 1, 'index': 26}],
        "conditional_resources": [
            {'item': 73,
             'resources': [{'amount': 5, 'index': 44},
                           {'amount': 1, 'index': 47},
                           {'amount': 1, 'index': 26}]}
        ],
        "convert": [],
    }


@pytest.mark.parametrize("simplified", [False, True])
def test_pickup_data_for_pb_expansion_locked(simplified, echoes_item_database, echoes_resource_database):
    # Setup
    pickup = pickup_creator.create_ammo_expansion(
        echoes_item_database.ammo["Power Bomb Expansion"],
        [2],
        True,
        echoes_resource_database,
    )
    if simplified:
        memo = claris_patcher_file._simplified_memo_data()
        hud_text = [
            "Power Bomb Expansion acquired, but the main Power Bomb is required to use it.",
            "Power Bomb Expansion acquired!",
        ]
    else:
        memo = default_database.default_prime2_memo_data()
        hud_text = [
            "Power Bomb Expansion acquired! \n"
            "Without the main Power Bomb item, you are still unable to release Power Bombs.",
            "Power Bomb Expansion acquired! \nMaximum Power Bomb carrying capacity increased by 2.",
        ]

    creator = pickup_exporter.PickupExporterSolo(memo)

    # Run
    details = creator.export(PickupIndex(0), PickupTarget(pickup, 0), pickup, PickupModelStyle.ALL_VISIBLE)
    result = claris_patcher_file.echoes_pickup_details_to_patcher(
        details, MagicMock())

    # Assert
    assert result == {
        "pickup_index": 0,
        "scan": "Power Bomb Expansion. Provides 2 Power Bombs and 1 Item Percentage",
        "model_name": "PowerBombExpansion",
        "hud_text": hud_text,
        'resources': [{'amount': 2, 'index': 72},
                      {'amount': 1, 'index': 47}],
        "conditional_resources": [
            {'item': 43,
             'resources': [{'amount': 2, 'index': 43},
                           {'amount': 1, 'index': 47}]}
        ],
        "convert": [],
    }


def test_pickup_data_for_pb_expansion_unlocked(echoes_item_database, echoes_resource_database):
    # Setup
    pickup = pickup_creator.create_ammo_expansion(
        echoes_item_database.ammo["Power Bomb Expansion"],
        [2],
        False,
        echoes_resource_database,
    )
    creator = pickup_exporter.PickupExporterSolo(claris_patcher_file._simplified_memo_data())

    # Run
    details = creator.export(PickupIndex(0), PickupTarget(pickup, 0), pickup, PickupModelStyle.ALL_VISIBLE)
    result = claris_patcher_file.echoes_pickup_details_to_patcher(
        details, MagicMock())

    # Assert
    assert result == {
        "pickup_index": 0,
        "scan": "Power Bomb Expansion. Provides 2 Power Bombs and 1 Item Percentage",
        "model_name": "PowerBombExpansion",
        "hud_text": ["Power Bomb Expansion acquired!"],
        'resources': [{'amount': 2, 'index': 43},
                      {'amount': 1, 'index': 47}],
        "conditional_resources": [],
        "convert": [],
    }


@pytest.mark.parametrize("disable_hud_popup", [False, True])
def test_create_pickup_all_from_pool(echoes_resource_database,
                                     default_layout_configuration,
                                     disable_hud_popup: bool
                                     ):
    item_pool = pool_creator.calculate_pool_results(default_layout_configuration,
                                                    echoes_resource_database)
    index = PickupIndex(0)
    if disable_hud_popup:
        memo_data = claris_patcher_file._simplified_memo_data()
    else:
        memo_data = default_prime2_memo_data()
    creator = pickup_exporter.PickupExporterSolo(memo_data)

    for item in item_pool.pickups:
        data = creator.export(index, PickupTarget(item, 0), item, PickupModelStyle.ALL_VISIBLE)
        for hud_text in data.hud_text:
            assert not hud_text.startswith("Locked")


def test_run_validated_hud_text():
    # Setup
    rng = MagicMock()
    rng.randint.return_value = 0
    details = pickup_exporter.ExportedPickupDetails(
        index=PickupIndex(0),
        scan_text="scan",
        hud_text=["Energy Transfer Module acquired!"],
        conditional_resources=[
            ConditionalResources(None, None, ()),
        ],
        conversion=[],
        model=PickupModel(
            game=RandovaniaGame.METROID_PRIME_ECHOES,
            name="EnergyTransferModule",
        ),
        other_player=False,
    )

    # Run
    data = claris_patcher_file.echoes_pickup_details_to_patcher(
        details, rng)

    # Assert
    assert data['hud_text'] == ['Run validated!']


@pytest.mark.parametrize("stk_mode", SkyTempleKeyHintMode)
@patch("randovania.games.patchers.claris_patcher_file._logbook_title_string_patches", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.hints.create_hints", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.sky_temple_key_hint.hide_hints", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.sky_temple_key_hint.create_hints", autospec=True)
@patch("randovania.games.patchers.claris_patcher_file._akul_testament_string_patch", autospec=True)
def test_create_string_patches(
                               mock_akul_testament: MagicMock,
                               mock_stk_create_hints: MagicMock,
                               mock_stk_hide_hints: MagicMock,
                               mock_item_create_hints: MagicMock,
                               mock_logbook_title_string_patches: MagicMock,
                               stk_mode: SkyTempleKeyHintMode,
                               mocker,
                               ):
    # Setup
    game = MagicMock()
    all_patches = MagicMock()
    rng = MagicMock()
    mock_item_create_hints.return_value = ["item", "hints"]
    mock_stk_create_hints.return_value = ["show", "hints"]
    mock_stk_hide_hints.return_value = ["hide", "hints"]
    player_config = PlayersConfiguration(0, {0: "you"})
    mock_logbook_title_string_patches.return_values = []
    mock_akul_testament.return_values = []
    area_namers = MagicMock()

    # Run
    result = claris_patcher_file._create_string_patches(
        HintConfiguration(sky_temple_keys=stk_mode),
        game,
        all_patches,
        area_namers,
        player_config,
        rng,
    )

    # Assert
    expected_result = ["item", "hints"]
    mock_item_create_hints.assert_called_once_with(all_patches, player_config, game.world_list, area_namers, rng)
    mock_logbook_title_string_patches.assert_called_once_with()
    mock_akul_testament.assert_called_once_with()

    if stk_mode == SkyTempleKeyHintMode.DISABLED:
        mock_stk_hide_hints.assert_called_once_with()
        mock_stk_create_hints.assert_not_called()
        expected_result.extend(["hide", "hints"])

    else:
        mock_stk_create_hints.assert_called_once_with(all_patches, player_config, game.resource_database,
                                                      area_namers, stk_mode == SkyTempleKeyHintMode.HIDE_AREA)
        mock_stk_hide_hints.assert_not_called()
        expected_result.extend(["show", "hints"])

    assert result == expected_result


def test_create_claris_patcher_file(test_files_dir):
    # Setup
    description = LayoutDescription.from_file(test_files_dir.joinpath("log_files", "seed_a.rdvgame"))
    player_index = 0
    preset = description.permalink.get_preset(player_index)
    cosmetic_patches = EchoesCosmeticPatches()

    # Run
    result = claris_patcher_file.create_patcher_file(description, PlayersConfiguration(player_index, {0: "you"}),
                                                     cosmetic_patches)

    # Assert
    assert isinstance(result["spawn_point"], dict)

    assert isinstance(result["pickups"], list)
    assert len(result["pickups"]) == 119

    assert isinstance(result["elevators"], list)
    assert len(result["elevators"]) == 22

    assert isinstance(result["translator_gates"], list)
    assert len(result["translator_gates"]) == 17

    assert isinstance(result["string_patches"], list)
    assert len(result["string_patches"]) == 61

    assert result["specific_patches"] == {
        "hive_chamber_b_post_state": True,
        "intro_in_post_state": True,
        "warp_to_start": preset.configuration.warp_to_start,
        "credits_length": 75 if cosmetic_patches.speed_up_credits else 259,
        "disable_hud_popup": cosmetic_patches.disable_hud_popup,
        "pickup_map_icons": cosmetic_patches.pickup_markers,
        "full_map_at_start": cosmetic_patches.open_map,
        "dark_world_varia_suit_damage": preset.configuration.varia_suit_damage,
        "dark_world_dark_suit_damage": preset.configuration.dark_suit_damage,
        "always_up_gfmc_compound": True,
        "always_up_torvus_temple": True,
        "always_up_great_temple": False,
    }
    # TODO: check all fields?
    assert result["dol_patches"]["default_items"] == {
        "visor": "Combat Visor",
        "beam": "Power Beam",
    }
